// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// bvector.cc
//
// 96-4-7:  Fixed range selection for none zero ls.

#include <stdlib.h> // strtoul()
#include <iostream.h>
#include "common/bvector.h"


// Value representing "pass the end" position.

bit_vector::position_type bit_vector::npos = (position_type) -1;

// Direct compare table.
//
// 0 ... everything else
// 1 ... l = r
// 2 ... l < r
// 3 ... l > r
//             R
//   COMP   0 1 X Z .(NV)
//          ---------
//       0| 1 2 0 0 0
//       1| 3 1 0 0 0
//   L   X| 0 0 0 0 0
//       Z| 0 0 0 0 0
//       .| 0 0 0 0 0

enum { COMPERR = 0, COMPEQU = 1, COMPLT = 2, COMPGT = 3 };
static int COMPTABLE[][5] =
	{
	{1,2,0,0,0},
	{3,1,0,0,0},
	{0,0,0,0,0},
	{0,0,0,0,0},
	{0,0,0,0,0}
	};


// Local functions, used only in this file.


static bit_vector::amount_type
find_hw(bit_vector::decimal_type val, bit_vector::decimal_type &wt)
	{
	// Find the highest weight of <val>.  Return the number of digits
	// in <wt>.
	wt = 1;
	bit_vector::amount_type i = 1;
	for (; (val/wt) >= 10; ++i)
		wt *= 10;
	return i;
	}


// sub_bit_vector.  These functions are here instead of in the header file
// becuase we need to have class bit_vector defined first.  The only way
// to have this was to move these member functions out of the header file.

bit_vector::sub_bit_vector::sub_bit_vector(bit_vector &bv,
		const position_type ms, const position_type ls)
	: _BV(bv), _begin(-1), _end(-2), _size(0)
	{
	// According to the Verilog LRM, <ms> and <ls> must be non-negative
	// and <ms> must reference a more significant bit than <ls>
	if (bv._msb_lower == false && ms >= ls)
		{
		_begin = ls;
		_end = ms;
		_size = (_end - _begin) + 1;
		}
	else if (bv._msb_lower == true && ms <= ls)
		{
		// When <msb_lower> is true, it means this register was
		// declared with [0:7].  Thus, when range selecting, we
		// must convert to the internal bit_vector representation.
		// This only affects range selection, and not logic or
		// arithmetic operations.
		position_type tmp = bv._end + bv._begin; // Inclusive indices.
		_begin = tmp - ls;
		_end = tmp - ms;
		_size = (_end - _begin) + 1;
		}
	else
		return;

	// Check for out of range condition.
	if (_begin < bv._begin || _end > bv._end)
		{
		_begin = -1;
		_end = -2;
		_size = 0;
		}
	}

inline
bit_vector::sub_bit_vector::sub_bit_vector(const sub_bit_vector &sbv)
	: _BV(sbv._BV), _begin(sbv._begin), _end(sbv._end), _size(sbv._size)
	{}

bit_vector::sub_bit_vector &
bit_vector::sub_bit_vector::operator=(const sub_bit_vector &sbv)
	{
	// Try to copy over what we need, only.
	position_type i = _begin;
	position_type j = sbv._begin;
	size_type k = 0;
	for (; k < _size ; ++i, ++j, ++k)
		{
		// Did we go past the end of <sbv>?  I.e <sbv._size> < <_size>
		if (j > sbv._end)
			break;
		// If we are unused, don't modify it.
		else if (_BV._bits[i] == NVL)
			continue;
		// Do not copy over unused bits.  Make them DCs in our
		// copy.  Is this the correct behavior?
		if (sbv._BV._bits[j] == NVL)
			_BV._bits[i] = DC;
		else
			_BV._bits[i] = sbv._BV._bits[j];
		}
	// Did we get everything we asked for?
	if (k < _size)
		memset(_BV._bits + k, LO, _size - k);
	return *this;
	}

bit_vector::sub_bit_vector &
bit_vector::sub_bit_vector::operator=(const bit_vector &bv)
	{
	// Exactly like the above.
	position_type i = _begin;
	position_type j = bv._begin;
	size_type k = 0;
	for (; k < _size ; ++i, ++j, ++k)
		{
		if (j > bv._end)
			break;
		else if (_BV._bits[i] == NVL)
			continue;
		if (bv._bits[j] == NVL)
			_BV._bits[i] = DC;
		else
			_BV._bits[i] = bv._bits[j];
		}
	if (k < _size)
		memset(_BV._bits + k, LO, _size - k);
	return *this;
	}

bit_vector::ostream_type &
bit_vector::sub_bit_vector::info(bit_vector::ostream_type &s) const
	{
	if (_BV._msb_lower)
		s << "BV[" << _begin << ":" << _end << "](" << _size << ")";
	else
		s << "BV[" << _end << ":" << _begin << "](" << _size << ")";
	return s;
	}

bit_vector::ostream_type &
operator<<(bit_vector::ostream_type &s, const bit_vector::sub_bit_vector &sbv)
	{
	// Since bits are kept least significant to most significant.
	// We do not need to reverse the order.
	bit_vector::position_type i = sbv._end;
	for (; i >= sbv._begin; --i)
		s << char (sbv._BV[i]);
	return s;
	}


// Public member functions for bit_vector class.


bit_vector::bit_vector(logic_type b)
	: _msb_lower(false)
	{
	// The default constructor.  If no default logic value is given,
	// then we assume the caller wants a invalid bit vector.
	if (b == NVL)
		{
		_begin = -1;
		_end = -2;
		_size = 0;
		_bits = 0;
		}
	else
		{
		_begin = _end = 0;
		_size = 1;
		_bits = new logic_type[1];
		_bits[0] = b;
		}
	}

bit_vector::bit_vector(bool num)
	: _msb_lower(false), _begin(0), _end(0), _size(1)
	{
	// Convert from bool type.
	_bits = new logic_type[_size];
	_bits[0] = (num == false) ? LO : HI;
	}

bit_vector::bit_vector(unsigned long value)
	: _msb_lower(false), _begin(0)
	{
	// Automatic conversion from integer.
	if (value == 0)
		{
		// A bit_vector of 0,base10, is a bit of value 0.
		_end = 0;
		_size = 1;
		_bits = new logic_type[1];
		_bits[0] = LO;
		return;
		}

	// We don't know how big the number is.  So allocate enough to
	// store the largest value.  Then copy only what we need later.
	position_type largest = 8 * sizeof(decimal_type);
	bit_vector tmp(largest-1, 0);
	decimal_type rem = value;
	position_type i = 0;
	for(; rem > 0 && i < largest; ++i)
		{
		// Now convert integer to binary.
		tmp._bits[i] = ((rem % 2) == 1) ? HI : LO;
		rem /= 2;
		}
	_size = i;
	_end = _size - 1;
	_bits = new logic_type[_size];
	memcpy(_bits, tmp._bits, _size);
	}

bit_vector::bit_vector(position_type ms, position_type ls, logic_type val)
	{
	// Create a bit vector with the specified size.
	// According to the Verilog LRM, <ls> can be greater than
	// <ms>.  Thus, we must be careful about the <begin>/<end>.
	if (ms < ls)
		{
		_msb_lower = true;
		_begin = ms;
		_end = ls;
		}
	else
		{
		_msb_lower = false;
		_begin = ls;
		_end = ms;
		}
	_size = (_end - _begin) + 1;
	_bits = new logic_type[_end + 1]; // <_begin> + <_size>!
	// If we begin offset from zero, we need to setup the unused bits.
	if (_begin > 0)
		memset(_bits, NVL, _begin);
	// Initialize array to <val>, default is don't-care.
	memset(_bits + _begin, val, _size);
	}

bit_vector::bit_vector(const str_type &str, base_type base, size_type len)
	: _msb_lower(false), _begin(0)
	{
	// Set our bit vector from the base specified in <base>, with
	// the value contained in <str>.  <len> is optional, and set
	// to the needed size.
	// Determine our length.
	if (len > 0)
		_size = len;
	else
		{
		_size = str.length();
		// Convert to bit size.
		switch(base)
			{
			case 2: break;
			case 8: _size *= 3; break;
			// We don't know the size for decimal.  Need to
			// calculate.  So let the conversion function do
			// it.  But, we still need to allocate something
			// for the memset.
			case 10: _size = 1; break;
			case 16: _size *= 4; break;
			}
		}
	_end = _size - 1;
	_bits = new logic_type[_size];
	if (str.data() != 0)
		{
		// Zero out the bits.
		memset(_bits, LO, _size);
		// Now, convert to binary.
		switch(base)
			{
			case 2: assign_from_bin(str); break;
			case 8: assign_from_oct(str); break;
			case 10: assign_from_dec(str); break;
			case 16: assign_from_hex(str); break;
			default:
				// <begin> is zero, no need add.
				memset(_bits, DC, _size);
				break;
			}
		}
	else
		memset(_bits, DC, _size);
	}

bit_vector::bit_vector(const bit_vector &bv)
	: _msb_lower(bv._msb_lower), _begin(bv._begin), _end(bv._end),
		_size(bv._size)
	{
	// Copy constructor, since this is used automatically by the
	// compiler, we should not do any special operations on the
	// vector.  So we do an exact copy.
	if (_size > 0)
		{
		// Copy constructor copies everything.  Note: <size> might be
		// smaller than the actual length of <bits>.
		_bits = new logic_type[_end+1];
		memcpy(_bits, bv._bits, _end+1);
		}
	else
		_bits = 0;
	}

bit_vector::bit_vector(const sub_bit_vector &sbv)
	: _msb_lower(false), _begin(0), _size(sbv._size)
	{
	// Make an exact copy like the copy constructor, except only the
	// specified range is taken into account.
	if (_size > 0)
		{
		_end = _size - 1;
		_bits = new logic_type[_size];
		memcpy(_bits, sbv._BV._bits + sbv._begin, _size);
		}
	else
		_bits = 0;
	}

bit_vector::~bit_vector()
	{
	// Free allocated memory.
	if (_bits != 0)
		delete [] _bits;
	}

bit_vector &
bit_vector::operator=(const bit_vector &bv)
	{
	// Assignment is not like copy constructor.  Assignment only wants
	// enough data to fit in our array.  If the right-hand-side is
	// larger, we will truncate it.  Else if it's smaller, we prepend
	// with zeros.  In the special case that we are invalid, then copy
	// everything.
	if (bv._size == 0)
		; // Has no effect.
	else if (_size == 0)
		{
		// If we were nothing before, copy everything over now.
		_msb_lower = bv._msb_lower;
		_begin = bv._begin;
		_end = bv._end;
		_size = bv._size;
		// For
		//	reg [8:1] a;
		// size does not show the actual length.
		_bits = new logic_type[_end+1];
		memcpy(_bits, bv._bits, _end+1);
		}
	else if (bv._size < _size)
		{
		memset(_bits + _begin, LO, _size);
		memcpy(_bits + _begin, bv._bits + bv._begin, bv._size);
		}
	else
		memcpy(_bits + _begin, bv._bits + bv._begin, _size);
	return *this;
	}

bit_vector &
bit_vector::operator=(const sub_bit_vector &sbv)
	{
	// Same as above.
	if (sbv._size == 0)
		; // Has no effect.
	else if (_size == 0)
		{
		_msb_lower = false;
		_begin = 0;
		_size = sbv._size;
		_end = _size - 1;
		_bits = new logic_type[_size];
		memcpy(_bits, sbv._BV._bits + sbv._begin, _size);
		}
	else if (sbv._size < _size)
		{
		memset(_bits, LO, _size);
		memcpy(_bits, sbv._BV._bits + sbv._begin, _size);
		}
	else
		memcpy(_bits, sbv._BV._bits + sbv._begin, _size);
	return *this;
	}

bit_vector &
bit_vector::operator&=(const bit_vector &bv)
	{
	// Logically AND <bv> to our bit vector.
	bit_vector::position_type start = bv._begin;
	bit_vector::position_type stop = bv._end;
	bit_vector::position_type i = _begin;
	for (; i <= _end && start <= stop; ++i, ++start)
		_bits[i] = _bits[i] & bv._bits[start];
	// We have more bits than <bv>.
	if (i <= _end)
		memset(_bits + i, LO, (_end + 1) - i);
	return *this;
	}

bit_vector &
bit_vector::operator|=(const bit_vector &bv)
	{
	// Logically OR <bv> to our bit vector.
	bit_vector::position_type start = bv._begin;
	bit_vector::position_type stop = bv._end;
	bit_vector::position_type i = _begin;
	for (; i <= _end && start <= stop; ++i, ++start)
		_bits[i] = _bits[i] | bv._bits[start];
	return *this;
	}

bit_vector &
bit_vector::operator^=(const bit_vector &bv)
	{
	// Logically XOR <bv> to our bit vector.
	bit_vector::position_type start = bv._begin;
	bit_vector::position_type stop = bv._end;
	bit_vector::position_type i = _begin;
	for (; i <= _end && start <= stop; ++i, ++start)
		_bits[i] = _bits[i] ^ bv._bits[start];
	// We have more bits than <bv>.
	if (i <= _end)
		{
		for (; i < _end; ++i)
			_bits[i] = _bits[i] ^ LO;
		}
	return *this;
	}

bit_vector::logic_type &
bit_vector::operator()(position_type idx)
	{
	// This operator is used by external functions, thus it has test
	// conditions to make sure there are no out of range accesses.
	static logic_type errbit = DC;
	// If <idx> is out of range.  We must return a DC.
	if (idx < _begin || idx > _end)
		return (errbit = DC);
	if (_msb_lower == true)
		return _bits[_end - idx];
	return _bits[idx];
	}

bit_vector::sub_bit_vector
bit_vector::operator()(position_type ms, position_type ls)
	{
	// Even though <ms> and <ls> must be non-negative and <ms> greater
	// than <ls>, we must return something.  So we let the constructor
	// do the testing and return whatever the constructor constructed.
	return sub_bit_vector(*this, ms, ls);
	}

bit_vector::operator bool() const
	{
	// Implicit conversion to bool for comparisons.
	bit_vector tmp(LO);
	return (*this == tmp) ? false : true;
	}

bit_vector::operator unsigned long() const
	{
	// Convert to integer for quicker arithmetic.  The result is
	// truncated to the size of unsigned long.
	unsigned long res = 0;
	position_type numbits = 8 * sizeof(res);
	for (position_type i = _begin; i < numbits && i <= _end; ++i)
		{
		if (_bits[i] > HI)
			return 0;
		res += int(_bits[i]) * (1 << i);
		}
	return res;
	}

const bit_vector::str_type
bit_vector::to_string(base_type base, size_type len) const
	{
	// Convert our bit vector into a printable string.  No special
	// characters are prepended to signify base type.  If any of the
	// bits in our vector are don't-cares or invalid, we set the rest
	// of the string to 'x's.
	// <Rnd> rounds the size given in <s> to the next higher size of the
	// base given in <b>.  I.e.  if s = 31, and b = 3 (octal), then the
	// next higher size is 33.  Thus, <Rnd> will evaluate to 33.
#define Rnd(s,b)	((s%b)==0 ? s : s + (b-s%b))
	int tmp[17] = { -1,-1,1,-1,-1,-1,-1,-1,3,-1,-1,-1,-1,-1,-1,-1,4 };

	// If the base is decimal (10), call the convert function for it.
	if (base == BASE10)
		return binary_2_decimal(len);
	else if (base > BASE16 || tmp[base] < 1)
		return str_type((len > 0) ? len : 1, 'x');

	// Find out the size of the resulting string.
	size_type size;
	if (len > 0)
		size = len;
	else
		size = Rnd(_size, tmp[base]) / tmp[base];
	if (size <= 0)
		return str_type(1, 'x');

	// For base 2, 8 and 16, the conversion process is almost the same.
	str_type res(size, '0');
	int pad = -1;
	position_type j = _begin;
	for (position_type i = size - 1; j <= _end && i != npos; --i)
		{
		switch(base)
			{
			case BASE2: pad = get_bits_as_int(j, 1); break;
			case BASE8: pad = get_bits_as_int(j, 3); break;
			case BASE16: pad = get_bits_as_int(j, 4); break;
			default: return str_type(size, 'x');
			}
		// Convert to ASCII.
		if (pad >= 10 && pad <= 15)
			res[i] = (pad - 10) + 'a';
		else if (pad >= 0)
			res[i] = pad + '0';
		else
			res[i] = 'x';
		}
	return res;
#undef Rnd
	}


bool
bit_vector::compare_x(const bit_vector &bv) const
	{
	// Compare bit vectors with don't-cares matching anything.
	if (_size >= bv._size)
		{
		bit_vector tmp(_size-1, 0);
		tmp = bv;
		return this->bit_wise_compare_x(tmp);
		}
	else
		{
		bit_vector tmp(bv._size-1, 0);
		tmp = *this;
		return tmp.bit_wise_compare_x(bv);
		}
	}

bool
bit_vector::compare_z(const bit_vector &bv) const
	{
	// Compare bit vectors with high impedance matching anything.
	if (_size >= bv._size)
		{
		bit_vector tmp(_size-1, 0);
		tmp = bv;
		return this->bit_wise_compare_z(tmp);
		}
	else
		{
		bit_vector tmp(bv._size-1, 0);
		tmp = *this;
		return tmp.bit_wise_compare_z(bv);
		}
	}


// Convenient friend functions for bit_vector class.


bool
operator==(const bit_vector &l, const bit_vector &r)
	{
	// Exact match comparison.
	if (l._size == r._size)
		return (memcmp(l._bits+l._begin, r._bits+r._begin, l._size) == 0);
	else if (l._size < r._size)
		{
		bit_vector tmp(r._size-1, 0);
		tmp = l;
		return (memcmp(tmp._bits+tmp._begin, r._bits+r._begin, r._size) == 0);
		}
	else
		{
		bit_vector tmp(l._size-1, 0);
		tmp = r;
		return (memcmp(l._bits+l._begin, tmp._bits+tmp._begin, l._size) == 0);
		}
	}

bool
operator<(const bit_vector &l, const bit_vector &r)
	{
	// Determine if the value of <l> is less than the value of <r>.
	int ret = 0;
	if (l._size == r._size)
		ret = l.less_greater(r);
	else if (l._size < r._size)
		{
		bit_vector tmp(r._size-1, 0);
		tmp = l;
		ret = tmp.less_greater(r);
		}
	else
		{
		bit_vector tmp(l._size-1, 0);
		tmp = r;
		ret = l.less_greater(tmp);
		}
	return (ret == COMPLT);
	}

bool
operator!(const bit_vector &bv)
	{
	// Return true if <bv> is a bit vector with value zero.  We do that
	// by creating bit vector of value zero, and compare it with <bv>.
	return (bv == bit_vector(LO));
	}

bit_vector
operator~(const bit_vector &bv)
	{
	// Invert all bits in <bv>.  We use the invert operator for Logic.
	bit_vector res(bv._size - 1, 0);
	bit_vector::position_type j = 0;
	for (bit_vector::position_type i = bv._begin; i <= bv._end; ++i, ++j)
		res._bits[j] = ~(bv._bits[i]);
	return res;
	}

bit_vector
operator&(const bit_vector &l, const bit_vector &r)
	{
	// Logically AND two vectors.
	bit_vector res;
	if (l._size >= r._size)
		{
		res = bit_vector(l._size - 1, 0);
		res = r; // Extend the smaller bit vector.
		res &= l;
		}
	else
		{
		res = bit_vector(r._size - 1, 0);
		res = l;
		res &= r;
		}
	return res;
	}

bit_vector
operator|(const bit_vector &l, const bit_vector &r)
	{
	// Logically OR two vectors.
	bit_vector res;
	if (l._size >= r._size)
		{
		res = bit_vector(l._size - 1, 0);
		res = r; // Extend the smaller bit vector.
		res |= l;
		}
	else
		{
		res = bit_vector(r._size - 1, 0);
		res = l;
		res |= r;
		}
	return res;
	}

bit_vector
operator^(const bit_vector &l, const bit_vector &r)
	{
	// Logically XOR two vectors.
	bit_vector res;
	if (l._size >= r._size)
		{
		res = bit_vector(l._size - 1, 0);
		res = r; // Extend the smaller bit vector.
		res ^= l;
		}
	else
		{
		res = bit_vector(r._size - 1, 0);
		res = l;
		res ^= r;
		}
	return res;
	}

//  Table for adding two bit vectors:
//  Ci A B | Sum Co
//  ---------------
//   0 0 0 |  0  0
//   0 0 1 |  1  0
//   0 1 0 |  1  0
//   0 1 1 |  0  1
//   1 0 0 |  1  0
//   1 0 1 |  0  1
//   1 1 0 |  0  1
//   1 1 1 |  1  1

struct bv_at // Bit Vector Add Table.
	{
	bit_vector::logic_type sum;
	bit_vector::logic_type carry;
	};
static struct bv_at ADDTABLE[] =
	{
	{LO,LO},
	{HI,LO},
	{HI,LO},
	{LO,HI},
	{HI,LO},
	{LO,HI},
	{LO,HI},
	{HI,HI}
	};

bit_vector
operator+(const bit_vector &l, const bit_vector &r)
	{
	// The result of an addition supports the carryout bit.  But
	// if the user doesn't save this bit, it is destroyed upon
	// call to destructor.  I'm not sure if Verilog has this behavior.
	bit_vector::size_type size = (l._size > r._size) ? l._size : r._size;
	bit_vector res(size, 0, LO); // Usually <size>-1 length.
	int pad = 0;
	bit_vector::logic_type A, B;
	bit_vector::logic_type ci = LO;
	for (bit_vector::position_type i = 0; i <= res._end; ++i)
		{
		A = B = NVL;
		// Have we reached the end of one of our vectors?
		if (i > r._end)
			{
			// If so, are there still something to add?
			if (ci == LO)
				{
				if (l._end > r._end)
					memcpy(res._bits+i, l._bits+i, (l._end+1)-i);
				return res;
				}
			else
				B = LO;
			}
		if (i > l._end)
			{
			if (ci == LO)
				{
				if (r._end > l._end)
					memcpy(res._bits+i, r._bits+i, (r._end+1)-i);
				return res;
				}
			else
				A = LO;
			}
		// Now try to add the vectors.
		if (A == NVL)
			A = l._bits[i];
		if (B == NVL)
			B = r._bits[i];
		if (A > HI || B > HI) // ci > HI, never happens.
			{
			// A DC or Z found.  Set the rest of the result to DC's.
			memset(res._bits+i, DC, (res._end+1)-i);
			return res;
			}
		pad = (int(ci) << 2) | (int(B) << 1) | int(A);
		res._bits[i] = ADDTABLE[pad].sum;
		ci = ADDTABLE[pad].carry;
		}
	return res;
	}

bit_vector
operator-(const bit_vector &l, const bit_vector &r)
	{
	// Substraction assumes both vectors are unsigned and does not
	// take into account borrows.
	bit_vector::size_type size = (l._size > r._size) ? l._size : r._size;
	bit_vector One(HI), res(size-1, 0);
	bit_vector ltmp(size-1, 0), rtmp(size-1, 0);
	ltmp = l; // Extend our vectors to the correct size.
	rtmp = r;
	rtmp = ~rtmp; // 1's compliment.
	rtmp = rtmp + One; // 2's compliment.
	res = ltmp + rtmp;
	return res;
	}

bit_vector
operator<<(const bit_vector &bv, bit_vector::amount_type amt)
	{
	// Shift <bv> to the left by <amt>.  Prepend zeros to the
	// beginning of the result.
	bit_vector res(bv._size-1, 0, LO);
	if (amt > res._end) // Is this standard behavior?
		return res;
	memcpy(res._bits+amt, bv._bits+bv._begin, res._size-amt);
	return res;
	}

bit_vector
operator>>(const bit_vector &bv, bit_vector::amount_type amt)
	{
	// Shift <bv> to the right by <amt>.  Appending zeros to the
	// front.
	bit_vector res(bv._size-1, 0, LO);
	bit_vector::position_type start = bv._begin + amt;
	if (start > bv._end) // Is this standard behavior?
		return res;
	memcpy(res._bits, bv._bits+start, (bv._end+1)-start);
	return res;
	}

bit_vector::ostream_type &
bit_vector::info(bit_vector::ostream_type &s) const
	{
	if (_msb_lower)
		s << "BV[" << _begin << ":" << _end << "](" << _size << ")";
	else
		s << "BV[" << _end << ":" << _begin << "](" << _size << ")";
	return s;
	}

bit_vector::ostream_type &
operator<<(bit_vector::ostream_type &s, const bit_vector &bv)
	{
	// Since our representation is from least to most significant bit.
	// We have to reverse the displaying process to make it more natural
	// for humans to read.  Humans like the msb at the left-hand side,
	// thus we must start at the end and move backwards.
	if (bv._size > 0)
		{
		bit_vector::position_type i = bv._end;
		for (; i >= bv._begin; --i)
			s << char (bv._bits[i]);
		}
	else
		s << "?01xz.?"; // Invalid bit vector.
	return s;
	}


// Protected member functions of bit_vector class.


const bit_vector::str_type
bit_vector::binary_2_decimal(size_type len) const
	{
	// Convert from binary to decimal.
	decimal_type wt = 0, num = 0;
	position_type j = _begin;
	amount_type numbits = 8 * sizeof(decimal_type);
	for (amount_type i = 0; j <= _end && i < numbits; ++i, ++j)
		{
		if (_bits[j] > HI)
			return str_type(len > 0 ? len : 1, 'x');
		num += int(_bits[j]) * (1 << i);
		}

	// Convert decimal to string.
	size_type size = 0;
	if (len > 0)
		size = len;
	else
		size = find_hw(num, wt);
	str_type res(size, '\0');
	decimal_type tmp1 = 10, tmp2 = 0;
	for (position_type i = size - 1; i != npos; --i, tmp1 *= 10)
		{
		tmp2 = num % tmp1;
		num -= tmp2;
		// Shift the number down to less than 10.
		tmp2 = tmp2 / (tmp1 / 10);
		res[i] = tmp2 + '0';
		}
	return res;
	}

int
bit_vector::get_bits_as_int(position_type &idx, amount_type numbits) const
	{
	// Get the amount of bits specified by <numbits> and return the
	// integer equivalent.  <idx> should be in range.  Return -1 if
	// any of the bits are out of range.
	int ret = 0;
	logic_type tmp = NVL;

	// Make sure we have enough.
	position_type remaining = (_end + 1) - idx;
	if (remaining > numbits)
		remaining = numbits;

	switch (remaining)
		{
		case 4:
			tmp = _bits[idx + 3];
			if (tmp > HI)
				{
				ret = -1;
				break;
				}
			else if (tmp == HI)
				ret += 0x08;
		case 3:
			tmp = _bits[idx + 2];
			if (tmp > HI)
				{
				ret = -1;
				break;
				}
			else if (tmp == HI)
				ret += 0x04;
		case 2:
			// This case is never used, but needed anyway.
			tmp = _bits[idx + 1];
			if (tmp > HI)
				{
				ret = -1;
				break;
				}
			else if (tmp == HI)
				ret += 0x02;
		case 1:
			tmp = _bits[idx];
			if (tmp > HI)
				{
				ret = -1;
				break;
				}
			else if (tmp == HI)
				ret += 0x01;
		}
	idx += remaining;
	return ret;
	}

bool
bit_vector::assign_bits(position_type &i,
		logic_type A, logic_type B, logic_type C, logic_type D)
	{
	// Assign the provided bits to our bit vector.  <D> is an optional
	// parameter.
	if (i > _end) return false;
	_bits[i++] = A;
	if (i > _end) return false;
	_bits[i++] = B;
	if (i > _end) return false;
	_bits[i++] = C;
	if (i > _end) return false;
	if (D != NVL)
		_bits[i++] = D;
	return true;
	}

void
bit_vector::sign_extend(position_type last)
	{
	// Sign extend our bit vector.  Last must be within range.
	size_type len = _end - last;
	if (len > 0)
		{
		logic_type hb = _bits[last]; // High bit.
		memset(_bits+last, hb, len);
		}
	}

void
bit_vector::assign_from_bin(const str_type &str)
	{
	// Set our bit vector from a binary string.
	position_type i = str.length() - 1;
	for (position_type j = _begin; j <= _end && i != npos; ++j, --i)
		{
		switch(str[i])
			{
			case '0': _bits[j] = LO; break;
			case '1': _bits[j] = HI; break;
			case 'x':
			case 'X': _bits[j] = DC; break;
			case 'z':
			case 'Z': _bits[j] = Z; break;
			case '_': --j; break; // Skip this character.
			default: return;
			}
		}
	}

void
bit_vector::assign_from_oct(const str_type &str)
	{
	// Set our bit vector from an octal string.
	position_type i = str.length() - 1;
	for (position_type j = _begin; j <= _end && i != npos; --i)
		{
		switch(str[i])
			{
			case '0': if (!assign_bits(j, LO, LO, LO)) return; break;
			case '1': if (!assign_bits(j, HI, LO, LO)) return; break;
			case '2': if (!assign_bits(j, LO, HI, LO)) return; break;
			case '3': if (!assign_bits(j, HI, HI, LO)) return; break;
			case '4': if (!assign_bits(j, LO, LO, HI)) return; break;
			case '5': if (!assign_bits(j, HI, LO, HI)) return; break;
			case '6': if (!assign_bits(j, LO, HI, HI)) return; break;
			case '7': if (!assign_bits(j, HI, HI, HI)) return; break;
			case 'x':
			case 'X': if (!assign_bits(j, DC, DC, DC)) return; break;
			case 'z':
			case 'Z': if (!assign_bits(j, Z, Z, Z)) return; break;
			case '_': --j; break;
			default: return;
			}
		}
	}

void
bit_vector::assign_from_dec(const str_type &str)
	{
	// Set our bit vector from a decimal string.
	decimal_type value = 0;
	if (str.find_first_of("XxZz") != str_type::npos)
		{
		// Set entire bit vector to don't cares.
		memset(_bits+_begin, DC, _size);
		return;
		}
	else
		{
		value = strtoul(str.c_str(), 0, 10);
		if (value == 0)
			// We can return if <value> is zero, because <bv>
			// is already set to zero by the caller.
			return;
		}

	// We don't know how big the number is.  So allocate enough to
	// store the largest value.  Then copy only what we need later.
	position_type numbits = 8 * sizeof(decimal_type);
	bit_vector tmp(numbits - 1, 0);
	decimal_type rem = value;
	position_type i = 0;
	for(; rem > 0 && i < numbits; ++i)
		{
		// Now convert integer to binary.
		tmp._bits[i] = ((rem % 2) == 1) ? HI : LO;
		rem /= 2;
		}
	if (_bits != 0)
		delete [] _bits;
	_size = i;
	_end = _size - 1;
	_bits = new logic_type[_size];
	memcpy(_bits, tmp._bits, _size);
	}

void
bit_vector::assign_from_hex(const str_type &str)
	{
	// Set our bit vector from a hexadecimal string.
	position_type i = str.length() - 1;
	for (position_type j = _begin; j <= _end && i != npos; --i)
		{
		switch(str[i])
			{
			case '0': if (!assign_bits(j, LO, LO, LO, LO)) return; break;
			case '1': if (!assign_bits(j, HI, LO, LO, LO)) return; break;
			case '2': if (!assign_bits(j, LO, HI, LO, LO)) return; break;
			case '3': if (!assign_bits(j, HI, HI, LO, LO)) return; break;
			case '4': if (!assign_bits(j, LO, LO, HI, LO)) return; break;
			case '5': if (!assign_bits(j, HI, LO, HI, LO)) return; break;
			case '6': if (!assign_bits(j, LO, HI, HI, LO)) return; break;
			case '7': if (!assign_bits(j, HI, HI, HI, LO)) return; break;
			case '8': if (!assign_bits(j, LO, LO, LO, HI)) return; break;
			case '9': if (!assign_bits(j, HI, LO, LO, HI)) return; break;
			case 'a':
			case 'A': if (!assign_bits(j, LO, HI, LO, HI)) return; break;
			case 'b':
			case 'B': if (!assign_bits(j, HI, HI, LO, HI)) return; break;
			case 'c':
			case 'C': if (!assign_bits(j, LO, LO, HI, HI)) return; break;
			case 'd':
			case 'D': if (!assign_bits(j, HI, LO, HI, HI)) return; break;
			case 'e':
			case 'E': if (!assign_bits(j, LO, HI, HI, HI)) return; break;
			case 'f':
			case 'F': if (!assign_bits(j, HI, HI, HI, HI)) return; break;
			case 'x':
			case 'X': if (!assign_bits(j, DC, DC, DC, DC)) return; break;
			case 'z':
			case 'Z': if (!assign_bits(j, Z, Z, Z, Z)) return; break;
			case '_': --j; break;
			default: return;
			}
		}
	}

int
bit_vector::less_greater(const bit_vector &bv) const
	{
	// Compare the two bit vectors <this> and <bv>.  Both vectors must be
	// the same size.  See COMPTABLE for return values.
	bit_vector::position_type li = _end;
	bit_vector::position_type ri = bv.end_pos();
	for (; li >= _begin; --li, --ri)
		{
		if (COMPTABLE[int (_bits[li])][int (bv[ri])] != COMPEQU)
			return COMPTABLE[int (_bits[li])][int (bv[ri])];
		}
	return COMPEQU;
	}

bool
bit_vector::bit_wise_compare_x(const bit_vector &bv) const
	{
	// <bv> must be the same size as us.
	position_type i = _begin, j = bv._begin;
	for (; i <= _end; ++i, ++j)
		{
		if (::compare_x(_bits[i], bv[j]) == false)
			return false;
		}
	return true;
	}

bool
bit_vector::bit_wise_compare_z(const bit_vector &bv) const
	{
	// <bv> must be the same size as us.
	position_type i = _begin, j = bv._begin;
	for (; i <= _end; ++i, ++j)
		{
		if (::compare_x(_bits[i], bv[j]) == false)
			return false;
		}
	return true;
	}
