// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// bvector.h
//
// Bit vector class.  Needed to represent don't cares in our bit vector.
// This object is a general purpose number object.  All arithmetic and
// logic operations are done with bit_vectors.  All number representations
// in a Verilog description will be represented by a bit_vector.
//
// Things to note:
// (1)  <begin> and <end> are inclusive indices.  This means they are
// the actual beginning and ending indices.
// (2)  The basic unit of logic is the Logic type.  See Logic.h.
// (3)  Logic bits are internally kept least significant bit to
//      most significant bit.  Thus, a conversion is needed for
//      declarations where the ls_exp is larger than the ms_exp.
// (4)  When a decimal string (char *) is passed as input to the
//      constructor, the number must be less than <dec_t> size.
//      Otherwise, we can't store the temporary number.

#ifndef _BVECTOR_H
#define _BVECTOR_H

#include <string>
#include <utility>
#include "common/logic.h"

class ostream;

class bit_vector
	{
public:
	// Size of a decimal number.  This should be set to the largest integer
	// size that the machine can handle.
#ifndef DECTYPE
# if defined(__GNUG__) && !defined(__STDC__)
	typedef long long decimal_type;
# else
	typedef long decimal_type;
# endif
#else
	typedef DECTYPE decimal_type;
#endif

	// The base type is used in conversion functions.
	enum base_type { BASE2 = 2, BASE8 = 8, BASE10 = 10, BASE16 = 16 };

	// Size type is used as a magnitude quantity.
	typedef unsigned long size_type;

	// Amount is like size, except it can be negative.
	typedef signed long amount_type;

	// Position type is used for locations/indices.  We allow a negative
	// location for error detection.  I.e. out of range situations.
	typedef signed long position_type;

	// This is our basic logic type.
	typedef logic logic_type;

	// Retype some library types.
	typedef string str_type;
	typedef ostream ostream_type;

	// A partial bit vector representation.  Used for range assignments
	// or references.
	class sub_bit_vector
		{
		// Give bit_vector access to constructor.
		friend class bit_vector;
		friend ostream_type &operator<<(ostream_type &,
			const sub_bit_vector &);

		bit_vector &_BV; // The original bit vector we are part of.
		position_type _begin; // Starting position.
		position_type _end; // Stopping position.
		size_type _size; // Size of this sub-bit vector.

		// Only bit_vector can create a sub_bit_vector object.
		sub_bit_vector(bit_vector &,
			const position_type, const position_type);
		sub_bit_vector(const sub_bit_vector &);
	public:
		sub_bit_vector &operator=(const sub_bit_vector &);
		sub_bit_vector &operator=(const bit_vector &);
		size_type size() const
			{ return _size; }
		ostream_type &info(ostream_type &) const;

		// We don't need any other method, yet!
		};

	// These friends are prototyped here because they need the above
	// types to be declared.
	friend bool operator==(const bit_vector &, const bit_vector &);
	friend bool operator<(const bit_vector &, const bit_vector &);
	friend bool operator!(const bit_vector &);
	friend bit_vector operator~(const bit_vector &);
	friend bit_vector operator&(const bit_vector &, const bit_vector &);
	friend bit_vector operator|(const bit_vector &, const bit_vector &);
	friend bit_vector operator^(const bit_vector &, const bit_vector &);
	friend bit_vector operator+(const bit_vector &, const bit_vector &);
	friend bit_vector operator-(const bit_vector &, const bit_vector &);
	friend bit_vector operator<<(const bit_vector &, amount_type);
	friend bit_vector operator>>(const bit_vector &, amount_type);
	friend ostream_type &operator<<(ostream_type &, const bit_vector &);
	friend ostream_type &operator<<(ostream_type &,
		const bit_vector::sub_bit_vector &);

	bit_vector(logic_type = NVL);
	bit_vector(bool);
	bit_vector(unsigned long);
	bit_vector(position_type, position_type, logic_type = DC);
	bit_vector(const str_type &, base_type = BASE2, size_type = 0);
	bit_vector(const bit_vector &);
	bit_vector(const sub_bit_vector &);
	virtual ~bit_vector();
	bit_vector &operator=(const bit_vector &);
	bit_vector &operator=(const sub_bit_vector &);
	bit_vector &operator&=(const bit_vector &);
	bit_vector &operator|=(const bit_vector &);
	bit_vector &operator^=(const bit_vector &);
	// These two are usually used in assignments, so no need constants.
	logic_type &operator()(position_type);
	sub_bit_vector operator()(position_type, position_type);
	// Implicit converters.
	operator bool() const;
	operator unsigned long() const;

	// Public helper functions.
	const str_type to_string(base_type = BASE10, size_type = 0) const;
	bool compare_x(const bit_vector &) const;
	bool compare_z(const bit_vector &) const;
	size_type size() const
		{ return _size; }
	position_type start_pos() const
		{ return _begin; }
	position_type end_pos() const
		{ return _end; }
	ostream_type &info(ostream_type &) const;

protected:
	// Invalid position.
	static position_type npos;

	// Only used by us.  To reference a range of bit_vector, use operator().
	logic_type &operator[](position_type idx)
		{ return _bits[idx]; }
	typedef const logic_type const_logic_type;
	const_logic_type operator[](position_type idx) const
		{ return _bits[idx]; }

	// Used by to_string to do the work for decimal case.
	const str_type binary_2_decimal(size_type = 0) const;

	// Other utility functions.
	int get_bits_as_int(position_type &, amount_type) const;
	bool assign_bits(position_type &,
		logic_type, logic_type, logic_type, logic_type = NVL);
	void sign_extend(position_type);
	void assign_from_bin(const str_type &);
	void assign_from_oct(const str_type &);
	void assign_from_dec(const str_type &);
	void assign_from_hex(const str_type &);
	// Functions used in comparisons.
	int less_greater(const bit_vector &) const;
	bool bit_wise_compare_x(const bit_vector &) const;
	bool bit_wise_compare_z(const bit_vector &) const;

private:
	// Verilog allows both:
	//	reg [16:0] a;
	//	reg [0:15] b;
	// Therefore, we must keep track of which was used.  Msb is lower
	// for the second case.  Note:  this is only used in range access.
	// I.e. internal bit storage is from least to most significant bit.
	// So arithmetic and logic operations are not affected.  This is
	// mostly used for error checking.
	bool _msb_lower;
	position_type _begin;
	position_type _end;
	size_type _size;
	logic_type *_bits;
	};

#endif // _BVECTOR_H
