// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// error.cc
//
// Modularized error messages.  See error.h on how to add more error
// detection.

#include <iostream.h>
#include <stdlib.h>
#include "common/error.h"

// progname: filename[sim_lineno] -sim_errno- SIM_ERR_MESG (arg)
// Example:
//
//	vbs: test.v[101] -4- syntax error, unexpected symbol (xxx)

string sim_program_name; // Our name to the system.
string sim_file_name; // File name with error.
int sim_lineno = 0; // Line number of error.
SIM_ERR_NUM sim_errno; // Error value.
SIM_STATE sim_state = SS_NONE;

string SIM_ERR_MESG[] =
	{
	string("no errors or warnings"),

	// Errors related to symbol table.
	string("symbol table not initialized"),
	string("symbol table is full"),
	string("duplicate symbol in symbol table"),
	string("undefined symbol"),

	// Paser errors.
	string("unexpected symbol"),
	string("unexpected keyword"),
	string("unexpected end of file"),

	// Errors related to task and function calls.
	string("not a task or system task"),
	string("not a function or system function"),
	string("Wrong number of arguments"),
	string("missing type specifier in format string"),
	string("unknown type specifier in format string"),
	string("invalid symbol in format string"),
	string("no termination in format string"),
	string("no format string specified"),
	string("ports and declarations do not match"),
	string("invalid register declaration"),
	string("invalid wire declaration"),
	string("invalid io declaration"),
	string("incomplete net type"),

	// General Errors...
	string("construct not yet supported"),
	string("internal simulator error"),
	string("no top-level module found"),
	string("invalid range"),
	string("order of indices are wrong"),
	string("only one default case is allowed"),
	string("invalid type for this usage"),
	string("not an lvalue"),
	string("repeated operation")
	};

void
sim_perror(strstream &mesg)
	{
	string s(mesg.str(), mesg.pcount());
	sim_perror(s);
	}

void
sim_perror(const string &mesg)
	{
	cout << endl << sim_program_name << ": ";
	cout << sim_file_name << '[';
	if (sim_lineno > 0)
		cout << sim_lineno;
	cout << "] ";

	if (sim_errno != SE_NONE)
		cout << "-" << sim_errno;
	cout << "- " << SIM_ERR_MESG[sim_errno];

	// Display the message.
	cout << " (" << mesg << ")." << endl << endl;

	if (sim_state & SS_SIMULATE)
		cout << "End simulation...";
	else if (sim_state & SS_SETUP)
		cout << "End setup...";
	else if (sim_state & SS_COMPILE)
		cout << "End compilation...";
	else
		cout << "End unknown state...";
	cout << endl;
	exit(sim_errno);
	}

void
sim_seterror(SIM_ERR_NUM num, const string &fn, int ln)
	{
	sim_errno = num;
	sim_file_name = fn;
	sim_lineno = ln;
	}

