// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// error.h
//
// This is the standard error reporting mechanism for VBS.  Any error
// conditions that vbs detects will be enumerated here as a macro.  If
// the error is not here, it's most likely not detected.  In the future,
// I hope to make this object streamable.
//
// To add a new error to detect, include a message in SIM_ERR_MESG
// (error.cc), and add a define in SIM_ERR_NUM.

#ifndef _ERROR_H
#define _ERROR_H

#include <strstream.h>
#include <string>

enum SIM_STATE
	{
	SS_NONE = 0x00,
	SS_COMPILE = 0x10,
	SS_SIMULATE = 0x20,
	SS_SETUP = 0x40
	};

enum SIM_ERR_NUM
	{
	SE_NONE, // No errors or warnings.

	// Errors related to symbol table.
	SE_STINIT, // Symbol table not initialized (if changed, change
	           // hash_value too).
	SE_STFULL, // Symbol table is full.
	SE_STDUP, // Duplicate symbol in symbol table.
	SE_STDEF, // Symbol not found.

	// Parser errors.
	SE_PSYMBOL, // Unexpected symbol.
	SE_PKEYWORD, // Unexpected keyword.
	SE_PEOF, // Unexpected end of file.

	// Errors related to task/function calls and module instantiations.
	SE_NTASK, // Not a task or system task.
	SE_NFUNCTION, // Not a function or system function.
	SE_NARGLST, // Wrong number of arguments.
	SE_FMTSPEC, // Missing type specifier in format string.
	SE_FMTUNKN, // Unknown type specifier in format string.
	SE_FMTSYM, // Invalid symbol in format string.
	SE_FMTNOTERM, // No termination in format string.
	SE_NOFMT, // No format string specified.
	SE_PORT, // Ports and declarations do not match.
	SE_REGDECL, // Invalid register declaration.
	SE_WIREDECL, // Invalid wire declaration.
	SE_IODECL, // Invalid io declaration.
	SE_COMPLETE, // Incomplete net type.

	// General Errors...
	SE_SUPPORT, // Construct not yet supported.
	SE_INTERNAL, // Internal simulator error.
	SE_NTOPLVL, // No top-level module found.
	SE_RANGE, // Invalid Range.
	SE_IDXORDER, // Order of indices are wrong.
	SE_MULTDEF, // Only one default case is allowed.
	SE_TYPE, // Invalid type for this usage.
	SE_NLVAL, // Not an lvalue.
	SE_REPEAT // Repeated operation.
	};

extern string sim_program_name; // Our name to the system.
extern string sim_file_name; // File name with error.
extern int sim_lineno; // Line number of error.
extern SIM_ERR_NUM sim_errno; // Error value.
extern SIM_STATE sim_state; // State of simulator.

void sim_perror(const string &);
void sim_perror(strstream &);
void sim_seterror(SIM_ERR_NUM, const string &, int);

#endif // _ERROR_H
