// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// event.h
//
// Event queue class to support "always @" statements.
//
// 96-1-15:  Added support for STL's list container and objectized event
// and trigger types.
//
// 96-2-10:  Fixed support for level change (reduce to one bit).
//
// 96-4-18:  Added support for variable change, renamed above object.

#ifndef _EVENT_H
#define _EVENT_H

#include <iostream.h>
#include <list>

// TRIG_TYPE is an object to replace the behavior of an enumeration.
// We need special operators for trigger type, but enumerations are
// inefficient (takes too much space).  This object should attempt to
// take the least amount of memory the system can support.

struct TRIG_TYPE
	{
	enum
		{
		TRIG_NONE    = 0x00,
		TRIG_NEGEDGE = 0x01,
		TRIG_POSEDGE = 0x02,
		TRIG_BITCHNG = 0x04,
		TRIG_CHNG    = 0x08
		};

	explicit TRIG_TYPE(int t = TRIG_NONE)
		: _type(t) {}
	TRIG_TYPE(const TRIG_TYPE &t)
		: _type(t._type) {}
	TRIG_TYPE operator=(const TRIG_TYPE &t)
		{ _type = t._type; return *this; }

	char _type;
	};

const TRIG_TYPE TRIG_NONE(TRIG_TYPE::TRIG_NONE);
const TRIG_TYPE TRIG_NEGEDGE(TRIG_TYPE::TRIG_NEGEDGE);
const TRIG_TYPE TRIG_POSEDGE(TRIG_TYPE::TRIG_POSEDGE);
const TRIG_TYPE TRIG_BITCHNG(TRIG_TYPE::TRIG_BITCHNG);
const TRIG_TYPE TRIG_CHNG(TRIG_TYPE::TRIG_CHNG);

inline bool
operator==(const TRIG_TYPE l, const TRIG_TYPE r)
	{ return l._type == r._type; }

inline TRIG_TYPE
operator|(const TRIG_TYPE l, const TRIG_TYPE r)
	{ return TRIG_TYPE(l._type | r._type); }

inline TRIG_TYPE
operator&(const TRIG_TYPE l, const TRIG_TYPE r)
	{ return TRIG_TYPE(l._type & r._type); }

// Handler for an event.

template<class T>
struct event_queue_handler
	{
	// The argument to this function is the object in which to
	// perform some task on.  The reference is used, so we do
	// not need to make a copy of T.
	virtual void operator()(T *) const
		{ cout << "No event handler, using default!" << endl; }
	};


// Base class for all event types.  Provide a common interface.

class ostream;

template<class T>
class event_base
	{
	typedef TRIG_TYPE trig_type;
	typedef T::expr_type bit_type;
	typedef bit_type::num_type num_type;
	typedef ostream ostream_type;

public:
	// Event container, holds all data for simulation.
	struct event_container
		{
		enum EVENT_STATE { ES_NON_VOLATILE, ES_VOLATILE, ES_USED };

		event_container(T *t)
			: _cnt(0), _queued(false), _event_state(ES_NON_VOLATILE), _obj(t)
			{}
		int _cnt; // Number of event with common object.
		bool _queued; // Were we queued already?
		EVENT_STATE _event_state;
		T *_obj; // Object to trigger.
		};

	event_base(event_container *c, const num_type &ival)
		: _previous_bit(ival), _bit(0) { _cache = c; _cache->_cnt++; }
	virtual ~event_base()
		{
		--(_cache->_cnt);
		if (_cache->_cnt == 0)
			{
			delete _cache;
			_cache = 0;
			}
		if (_bit != 0)
			delete _bit;
		}

	// The common interface.
	const num_type previous_bit() const
		{ return _previous_bit; }
	void previous_bit(const num_type &b)
		{ _previous_bit = b; }
	const bit_type *monitor_bit() const
		{ return _bit; }
	void monitor_bit(bit_type *b)
		{ _bit = b; }
	T *data()
		{ return _cache->_obj; }
	void queue()
		{ _cache->_queued = true; }
	void unqueue()
		{ _cache->_queued = false; }
	bool is_queued() const
		{ return (_cache->_queued == true); }
	event_container::EVENT_STATE event_state() const
		{ return _cache->_event_state; }
	void event_state(event_container::EVENT_STATE s)
		{ _cache->_event_state = s; }

	virtual ostream_type &display(ostream_type &s) const = 0;
	virtual trig_type type() const = 0;

private:
	num_type _previous_bit;
	bit_type *_bit; // Expression being monitored.  Evaluates to one bit.
	event_base<T>::event_container *_cache; // Common holding area.
	};

template<class T>
class bit_change_event : public event_base<T>
	{
public:
	bit_change_event(event_base<T>::event_container *c, const num_type &ival)
		: event_base<T>(c, ival) {}

	ostream_type &display(ostream_type &s) const
		{ s << "TRIG_BITCHNG" << endl; return s; }
	trig_type type() const
		{ return TRIG_BITCHNG; }
	};

template<class T>
class change_event : public event_base<T>
	{
public:
	change_event(event_base<T>::event_container *c, const num_type &ival)
		: event_base<T>(c, ival) {}

	ostream_type &display(ostream_type &s) const
		{ s << "TRIG_CHNG" << endl; return s; }
	trig_type type() const
		{ return TRIG_CHNG; }
	};

template<class T>
class posedge_event : public event_base<T>
	{
public:
	posedge_event(event_base<T>::event_container *c, const num_type &ival)
		: event_base<T>(c, ival) {}

	ostream_type &display(ostream_type &s) const
		{ s << "TRIG_POSEDGE" << endl; return s; }
	trig_type type() const
		{ return TRIG_POSEDGE; }
	};

template<class T>
class negedge_event : public event_base<T>
	{
public:
	negedge_event(event_base<T>::event_container *c, const num_type &ival)
		: event_base<T>(c, ival) {}

	ostream_type &display(ostream_type &s) const
		{ s << "TRIG_NEGEDGE" << endl; return s; }
	trig_type type() const
		{ return TRIG_NEGEDGE; }
	};

template<class T> inline ostream &
operator<<(ostream &s, const event_base<T> &d)
	{ return d.display(s); }


// Event queue used for the simulation timing.  The queue consists
// of a list of events (defined above).  A member function is provided
// to trigger all events in the list.  Usually, we want to trigger
// all events queued by the previous time unit.  Maybe I should
// provide simpler primitives for the user to trigger the events
// individually?


template<class T>
class event_queue
	{
public:
	typedef event_base<T>::trig_type trig_type;
	typedef event_base<T> event_type;
	typedef event_type *event_ptr;
	typedef list<event_ptr> event_list;

	void sim_add_event(event_ptr ev)
		{
		// We are about to put the event into our list.
		// We must change the queued status so we do not
		// try to queue it more than once.
		ev->queue();
		// Now append the event into the list.
		_event_list.push_back(ev);
		}
	void sim_trigger(const event_queue_handler<T> &h)
		{
		// Trigger events, but do not delete the event object.
		// It is not a copy, the actual event is stored in the
		// symbol table and should not be deleted.
		while (!_event_list.empty())
			{
			event_ptr ev = _event_list.front(); // Get event to trigger.
			h(ev->data()); // Trigger it.
			// Reset flag so we can queue this event next time.
			ev->unqueue();
			// Remove node from list.
			_event_list.pop_front();
			}
		}

private:
	event_list _event_list; // List of events to trigger.
	};

#endif // _EVENT_H
