// Verilog Behavioral Simulator
// Copyright (C) 1996, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// logic.h
//
// This object is the basic unit of binary representation (bit).
// It provides all the standard logic operations.  The size of the
// logic type is the same as sizeof(char).  This makes it possible
// to use memcpy/mem* functions to move data around.  By creating
// a class object, we have better type safety than if we simply
// used a typedef.  There are no overhead, since no members are
// virtual (no destructor).

#ifndef _LOGIC_H
#define _LOGIC_H

class logic
	{
	friend bool operator<(const logic, const logic);

public:
	enum state_value { LO, HI, DC, Z, NVL };

	logic(state_value l = NVL)
		: _state(l) { }
	logic(const logic &l)
		: _state(l._state) { }

	logic &operator=(const logic l)
		{ _state = l._state; return *this; }

	operator int() const
		{ return _state; }
	operator char() const
		{
		switch (_state)
			{
			case LO: return '0';
			case HI: return '1';
			case DC: return 'x';
			case  Z: return 'z';
			default: return '.';
			}
		}
protected:
	char _state; // Use the smallest type known.
	};

// Misc. operators.

// Constant values and tables for logic arithmetic.
// Currently, logic operations are through table lookup.  Can this be
// changed?  I want to make the tables static const members of
// logic.  But GCC produces 'multiple defined' link errors.  I guess
// I have to move these out of the header if I want to make them static.
// But that made things complicated, so I'm keeping everything in this
// header.

const logic  LO(logic::LO);
const logic  HI(logic::HI);
const logic  DC(logic::DC);
const logic   Z(logic::Z);
const logic NVL(logic::NVL); // Not Valid logic.

const int NUMLOGIC = 5;

// Boolean states of logic comparison.

const bool logic_EQ[][NUMLOGIC] =
	{ // Exact comparison.
	  {true,  false, false, false, false},
	  {false, true,  false, false, false},
	  {false, false, true,  false, false},
	  {false, false, false, true,  false},
	  {false, false, false, false, true},
	};

const bool logic_EQX[][NUMLOGIC] =
	{ // Compare with don't-care.
	  {true,  false, true,  false, false},
	  {false, true,  true,  false, false},
	  {true,  true,  true,  true,  false},
	  {false, false, true,  true,  false},
	  {false, false, false, false, true},
	};

const bool logic_EQZ[][NUMLOGIC] =
	{ // Compare with high impedance.
	  {true,  false, false, true,  false},
	  {false, true,  false, true,  false},
	  {false, false, true,  true,  false},
	  {true,  true,  true,  true,  false},
	  {false, false, false, false, true},
	};

// Logic operations.

const logic logic_INV[NUMLOGIC] =
	{ HI,LO,DC,DC,DC };

const logic logic_AND[][NUMLOGIC] =
	{ {LO,LO,DC,DC,DC},
	  {LO,HI,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	  {DC,DC,DC,DC,DC}
	};

const logic logic_OR[][NUMLOGIC] =
	{ {LO,HI,DC,DC,DC},
	  {HI,HI,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	  {DC,DC,DC,DC,DC}
	};

const logic logic_XOR[][NUMLOGIC] =
	{ {LO,HI,DC,DC,DC},
	  {HI,LO,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	  {DC,DC,DC,DC,DC},
	};

// Some logic arithmetic operations.

inline bool
compare_x(const logic l, const logic r)
	{ return logic_EQX[int (l)][int (r)]; }

inline bool
compare_z(const logic l, const logic r)
	{ return logic_EQZ[int (l)][int (r)]; }

inline bool
operator==(const logic l, const logic r)
	{ return logic_EQ[int (l)][int (r)]; }

inline bool
operator<(const logic l, const logic r)
	{ return l._state < r._state; }

inline bool
operator!(const logic l)
	{ return (l == LO) ? true : false; }

inline logic
operator~(const logic l)
	{ return logic_INV[int (l)]; }

inline logic
operator&(const logic l, const logic r)
	{ return logic_AND[int (l)][int (r)]; }

inline logic
operator|(const logic l, const logic r)
	{ return logic_OR[int (l)][int (r)]; }

inline logic
operator^(const logic l, const logic r)
	{ return logic_XOR[int (l)][int (r)]; }

#endif // _LOGIC_H
