// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// st_node.h
//
// STnode is a base class for all nodes in the symbol table.

#ifndef _ST_NODE_H
#define _ST_NODE_H

#include "common/bvector.h"
#include "common/hash.h"
#include "common/dstack.h"
#include "common/combase.h"

class st_function;
class st_instantiation;
class st_module;
class st_net;
class st_task;

struct st_node_monitor;
struct st_node_evaluate;
struct st_node_setup;
struct st_node_trigger;

struct st_node_base : public virtual common_base
	{
	typedef bit_vector num_type;
	typedef hash_value hash_type;
	typedef hash_type::scope_type scope_type;
	typedef Stack<int> scopelist_type;

	typedef st_node_monitor monitor_type;
	typedef st_node_evaluate evaluate_type;
	typedef st_node_setup setup_type;
	typedef st_node_trigger trigger_type;

	virtual ~st_node_base()
		{}

	// These functions must be defined by derived classes.
	virtual const str_type type_name() const = 0;
	virtual const str_type name() const = 0;
	virtual const scope_type scope() const = 0;
	virtual void scope(scope_type) = 0;

	// Use these to determine the node type.
	virtual st_function *get_function()
		{ return 0; }
	virtual st_instantiation *get_instantiation()
		{ return 0; }
	virtual st_module *get_module()
		{ return 0; }
	virtual st_net *get_net()
		{ return 0; }
	virtual st_task *get_task()
		{ return 0; }

	virtual ostream_type &display(ostream_type &s) const
		{ return s; }
	virtual bool sim_monitor(const monitor_type &) const
		{ return false; }
	virtual num_type sim_evaluate(const evaluate_type &) const
		{ return num_type(DC); }
	virtual void sim_setup(const setup_type &)
		{}
	virtual bool sim_trigger(const trigger_type &)
		{ return true; }
	};

inline ostream &
operator<<(ostream &s, const st_node_base &d)
	{ return d.display(s); }

struct st_node_monitor
	{
	virtual bool operator()(st_function *) const
		{ return false; }
	virtual bool operator()(st_net *) const
		{ return false; }
	};

struct st_node_evaluate
	{
	typedef st_node_base::num_type num_type;
	virtual num_type operator()(const st_function *) const
		{ return num_type(DC); }
	virtual num_type operator()(const st_net *) const
		{ return num_type(DC); }
	};

struct st_node_setup
	{
	virtual void operator()(st_function *) const {}
	virtual void operator()(st_instantiation *) const {}
	virtual void operator()(st_module *) const {}
	virtual void operator()(st_net *) const {}
	virtual void operator()(st_task *) const {}
	};

struct st_node_trigger
	{
	virtual bool operator()(st_task *) const { return true; }
	};

#endif // _ST_NODE_H
