// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// st_setup.cc

#include "common/error.h"
#include "common/sym_tab.h"
#include "common/st_mod.h"
#include "common/st_net.h"
#include "common/st_func.h"
#include "common/st_task.h"
#include "common/st_setup.h"
#include "stmt/ssetup.h"
#include "misc/msetup.h"
#include "moditm/tfsetup.h"
#include "moditm/mitfsetup.h"
#include "moditm/misetup.h"

extern symbol_table symboltable;

// sim_start

setup_st_node::setup_st_node(scopelist_type &s, port_conn_list *p, arg_list *a)
	: _scope(s)
	{ _port_conn = p; _caller_arg = a; }

void
setup_st_node::operator()(st_module *p) const
	{
	// Setup local scope.  Each new module has its own scope.
	int thisscope = symboltable.scope()++;
	scopelist_type localscope;
	localscope.push(0); /* Global scope. */
	localscope.push(thisscope); /* Local scope. */

	// Create ports first.  We have to create the symbol table nets before
	// any io declarations are done in the module items.
	if (p->_port_list != 0)
		{
		if (_port_conn == 0)
			{
			sim_seterror(SE_NARGLST, p->_filename, p->_lineno);
			sim_perror(p->_name);
			}
		else if (p->_port_list->size() != _port_conn->size())
			{
			sim_seterror(SE_NARGLST, p->_filename, p->_lineno);
			sim_perror(p->_name);
			}
		// Ports need port connections.  Pass the port connections to
		// setup the ports.
		port_conn_list::iterator idx(_port_conn->begin());
		module_type::port_list::iterator itp(p->_port_list->begin());
		module_type::port_list::iterator stop(p->_port_list->end());
		for (; itp != stop; ++idx, ++itp)
			// Ports and port connections list are the same length.
			// So no need worry about going pass the end.
			(*itp)->sim_setup(setup_port(localscope, (*idx).get()));
		// Create a list to store the io variables.
		p->_iovariable = new module_type::io_list;
		}
	if (p->_module_item != 0)
		{
		module_type::module_item_list::iterator itp(p->_module_item->begin());
		module_type::module_item_list::iterator stop(p->_module_item->end());
		for (; itp != stop; ++itp)
			(*itp)->sim_setup(setup_module_item(localscope, p->_iovariable));
		}

	// Now check to make sure the port connections and io declarations
	// match.
	if (p->_port_list != 0)
		{
		if (p->_port_list->size() != p->_iovariable->size())
			{
			sim_seterror(SE_PORT, p->_filename, p->_lineno);
			sim_perror(p->_name);
			}
		// Check to make sure the io variables are complete.
		module_type::io_list::iterator itp(p->_iovariable->begin());
		module_type::io_list::iterator stop(p->_iovariable->end());
		for (; itp != stop; ++itp)
			{
			net_type *net = symboltable.get(*itp)->get_net();
			if (net == 0)
				{
				sim_seterror(SE_INTERNAL, p->_filename, p->_lineno);
				sim_perror(p->_name);
				}
			else if (!net->complete())
				{
				sim_seterror(SE_COMPLETE, net->_filename, net->_lineno);
				sim_perror(net->_name);
				}
			}
		// Modules do not need IO variables, so get rid of it.
		// It was created just for this test.
		delete p->_iovariable;
		p->_iovariable = 0;
		}
	}

void
setup_st_node::operator()(st_net *p) const
	{
	if (!p->complete())
		{
		sim_seterror(SE_COMPLETE, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	}

SIM_ERR_NUM
check_iovariable(Stack<int> &s, list<hash_value> &i)
	{
	list<hash_value>::iterator idx(i.begin());
	list<hash_value>::iterator stop(i.end());
	for (; idx != stop; ++idx)
		{
		st_net *node = symboltable.get(*idx)->get_net();
		if (node == 0)
			return SE_TYPE;
		// This call checks the complete flag for nets.  See above.
		node->sim_setup(setup_st_node(s));
		}
	return SE_NONE;
	}

void
setup_st_node::operator()(st_function *p) const
	{
	// We do not need to setup the declaration and statement more than once.
	if (p->_done_setup != false)
		return;

	// Increment the scope for local data.  Declarations should not look
	// at parent scope, so do not include it.
	int thisscope = symboltable.scope()++;
	scopelist_type localscope;
	localscope.push(thisscope);

	// Must setup declarations first, because the statement might
	// need variables declared locally.
	if (p->_declaration != 0)
		{
		func_type::decl_list::iterator itp(p->_declaration->begin());
		func_type::decl_list::iterator stop(p->_declaration->end());
		func_type::io_list *iov = new func_type::io_list;
		for (; itp != stop; ++itp)
			(*itp)->sim_setup(setup_tfdecl(localscope, iov));
		// We no longer need this object.  But don't delete it,
		// it was not a copy.
		p->_declaration = 0;
		if (iov->size() != 0)
			p->_iovariable = iov;
		else
			{
			delete iov;
			iov = 0;
			}
		}

	// The scope for statements are:  global scope, parent scope, and
	// local scope.  Parent scope is needed for things like function
	// enables within the same module.
	localscope.pop(); // Move local scope to top of stack.
	localscope.push(0); // Global scope.
	localscope.push(_scope.top()); // Parent scope.
	localscope.push(thisscope);

	// The name of the function is the return value, so must setup
	// statement after entering the function name into the symbol
	// table.
	p->_stmt->sim_setup(setup_stmt(localscope));

	// Update flag so we do setup only once.
	p->_done_setup = true;
	}

void
setup_st_node::operator()(st_task *p) const
	{
	// We do not need to setup the declaration and statement more than once.
	if (p->_done_setup != false)
		return;

	// Increment the scope for local data.  Tasks do not look at global
	// scope.
	int thisscope = symboltable.scope()++;
	scopelist_type localscope;
	localscope.push(thisscope);

	// Must setup declarations first, because the statement might
	// need variables declared locally.
	if (p->_declaration != 0)
		{
		task_type::decl_list::iterator itp(p->_declaration->begin());
		task_type::decl_list::iterator stop(p->_declaration->end());
		task_type::io_list *iov = new task_type::io_list;
		for (; itp != stop; ++itp)
			(*itp)->sim_setup(setup_tfdecl(localscope, iov));
		// We no longer need this object.  But don't delete it,
		// it was not a copy.
		p->_declaration = 0;
		if (iov->size() != 0)
			p->_iovariable = iov;
		else
			{
			delete iov;
			iov = 0;
			}
		}

	// The scope for statements are:  global scope, parent scope, and
	// local scope.  Parent scope is needed for things like function
	// enables within the same module.
	localscope.pop(); // Move local scope to top of stack.
	localscope.push(0); // Global scope.
	localscope.push(_scope.top()); // Parent scope.
	localscope.push(thisscope);

	// Setup statement.
	p->_stmt->sim_setup(setup_stmt(localscope));

	// Update flag so we do setup only once.
	p->_done_setup = true;
	}

void
setup_st_node::operator()(st_instantiation *) const
	{
	// Nothing to do.
	}
