// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// st_util.cc

#include "common/error.h"
#include "common/hash.h"
#include "common/event.h"
#include "common/error.h"
#include "common/sym_tab.h"
#include "common/st_net.h"
#include "expr/exprbase.h"
#include "expr/rangeid.h"
#include "expr/eeval.h"
#include "stmt/stmtbase.h"

extern symbol_table symboltable;
extern event_queue<stmt_base> eventqueue;

// Utility functions for st_task/st_function's sim_evaluate and sim_setup
// and sim_trigger.

void
print_escapecode(const string &str, size_t &idx)
	{
	strstream buf;
	int i = 0;
	switch(str[idx])
		{
		case 'n': cout << endl; break;
		case 't': cout << '\t'; break;
		case '\\': cout << '\\'; break;
		case '\"': cout << '\"'; break;
		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
			while (str[idx] >= '0' && str[idx] <= '7')
				{
				buf << str[idx];
				++i;
				// Only 1-3 octal digits.
				if (i > 2)
					break;
				++idx;
				}
			// Now, convert the octal number to the equivalent
			// integer.  Then print out the ASCII value.
			cout << (char) strtol(buf.str(), 0, 8);
			break;
		default: cout << str[idx]; break;
		}
	}

bool
find_format(size_t &j, const string &fmt, string &str)
	{
	// Extract the format string from this format.
	// Return next unused position in <fmt> or -1 if error.
	strstream s;
	for (; j < fmt.size(); ++j)
		{
		if (fmt[j] == '\"')
			break;
		else if (fmt[j] >= '0' && fmt[j] <= '9')
			s << fmt[j];
		else if (fmt[j] == 'd' || fmt[j] == 'h'
				|| fmt[j] == 'b' || fmt[j] == 'o'
				|| fmt[j] == 's')
			{
			// Found the end.  These are our supported types.
			s << fmt[j];
			s << '\0';
			str = s.str();
			return true;
			}
		else if (fmt[j] == ' ')
			{
			// Save string for error reporting.
			s << '\0';
			str = s.str();
			sim_errno = SE_FMTSPEC;
			return false;
			}
		else
			{
			char tmp[2];
			tmp[0] = fmt[j];
			tmp[1] = '\0';
			str = tmp;
			sim_errno = SE_FMTSYM;
			return false;
			}
		}
	s << '\0';
	str = s.str();
	sim_errno = SE_FMTNOTERM;
	return false;
	}

bool 
check_arguments(list< basic_ptr<expr_base> > *t, list<hash_value> *tn)
	{
	// Check to make sure that the arguments exist if arguments are
	// taken by the function or task.
	// Return true if ok, false if something is wrong.
	if (tn == 0)
		{
		if (t != 0)
			return false;
		return true;
		}
	else
		{
		if (t == 0)
			return false;
		return tn->size() == t->size();
		}
	}

SIM_ERR_NUM
check_format(const string &fmt)
	{
	for (size_t i = 1; i < fmt.size(); ++i)
		{
		switch(fmt[i])
			{
			case '%':
				{
				string s;
				i++;	// Move to next character.
				if (!find_format(i, fmt, s))
					return sim_errno;
				break;
				}
			case '\\':
				break;
			default:
				break;
			}
		}
	return SE_NONE;
	}

void
entry_iovars(list<hash_value> &localvars, list< basic_ptr<expr_base> > &arg)
	{
	// Go through the list and initialize all input variables.
	list<hash_value>::iterator idx(localvars.begin());
	list<hash_value>::iterator last(localvars.end());
	list< basic_ptr<expr_base> >::iterator exp(arg.begin());
	st_net *node;
	for (; idx != last; ++idx, ++exp)
		{
		// Can not be null, since we created this node.
		node = symboltable.get(*idx)->get_net();
		// We do not use assignment here because this is not
		// an assignment.  We're just updating the local variables.
		if (node->_iodirection == st_net::IN
		 || node->_iodirection == st_net::INOUT)
			*node->_storage = (*exp)->sim_evaluate(evaluate_expr());
		}
	}

void
exit_iovars(list<hash_value> &localvars, list< basic_ptr<expr_base> > &arg)
	{
	// Go through the list and set all output variables.
	list<hash_value>::iterator idx(localvars.begin());
	list<hash_value>::iterator last(localvars.end());
	list< basic_ptr<expr_base> >::iterator exp(arg.begin());
	st_net *them, *us;
	for (; idx != last; ++idx, ++exp)
		{
		// Can not be null, since we created this node.
		us = symboltable.get(*idx)->get_net();
		if (us->_iodirection == st_net::OUT
		 || us->_iodirection == st_net::INOUT)
			{
			range_id *rid = (*exp)->get_range_id();
			// Can not be null.
			them = symboltable.get(rid->index())->get_net();
			// <ms>/<ls> are not used here because it's
			// upto <them> to select which bits to keep.
			// We just send them everything we have.
			them->assignment(*us->_storage);
			}
		}
	}


// Assignment function object for st_net.

void
check_monitor(list< counted_ptr< event_base<stmt_base> > > &el,
		bit_vector &before, bit_vector &after)
	{
	event_base<stmt_base> *ev = 0;
	st_net::monitor_list::iterator itp(el.begin());
	st_net::monitor_list::iterator stop(el.end());
	for (; itp != stop; ++itp)
		{
		bool queued = false;
		ev = (*itp).get();
		// Take care of pre-queuing procedure.  This is used to check
		// whether this event should be triggered, or was it triggered
		// by another event, and was used up.  This is the situation
		// for an event control in an initial procedural block that
		// monitors more than one variable.
		if (ev->event_state() == st_net::cache_type::ES_USED)
			{
			if (!ev->is_queued())
				{
				// Not queued, remove node.
				st_net::monitor_list::iterator delme(itp);
				--itp;
				el.erase(delme);
				}
			// An event marked as used should not be triggered again.
			continue;
			}
		if (ev->monitor_bit() != 0)
			{
			// We only care about a single bit.
			bit_vector bit(ev->monitor_bit()->sim_evaluate(evaluate_expr()));
			// Determine whether to queue event.
			if (!ev->is_queued())
				{
				if (bit == LO && ev->previous_bit() == HI)
					{
					// Not queued, queue it if the type is correct.
					event_base<stmt_base>::trig_type t = TRIG_CHNG
												| TRIG_BITCHNG
												| TRIG_NEGEDGE;
					if ((ev->type() & t) != TRIG_NONE)
						{
						eventqueue.sim_add_event(ev);
						queued = true;
						}
					}
				else if (bit == HI && ev->previous_bit() == LO)
					{
					event_base<stmt_base>::trig_type t = TRIG_CHNG
												| TRIG_BITCHNG
												| TRIG_POSEDGE;
					if ((ev->type() & t) != TRIG_NONE)
						{
						eventqueue.sim_add_event(ev);
						queued = true;
						}
					}
				}
			ev->previous_bit(bit);
			}
		else if (before != after)
			{
			// Determine whether to queue event.
			if (!ev->is_queued())
				{
				if ((ev->type() & TRIG_CHNG) != TRIG_NONE)
					{
					eventqueue.sim_add_event(ev);
					queued = true;
					}
				}
			}
		// Else, this variable didn't change.  Take care of post-
		// queuing procedure.
		if (queued && ev->event_state() == st_net::cache_type::ES_VOLATILE)
			// A volatile event is only used once, so mark it now.
			// We can not delete this event yet, since it was just
			// queued.
			ev->event_state(st_net::cache_type::ES_USED);
		}
	}

st_net_assign::st_net_assign(st_net &n)
	: _net(n)
	{}

void
st_net_assign::operator()(num_type &num, position_type m, position_type l)
	{
	num_type before(*_net._storage);
	if (m >= 0 && l < 0)
		(*_net._storage)(m, m) = num;
	else if (m >= 0 && l >= 0)
		(*_net._storage)(m, l) = num;
	else
		*_net._storage = num;

	// If we haven't changed due to the assignment, there's nothing
	// to do after this.  It is though this assignment never occured.
	num_type after(*_net._storage);
	if (before == after)
		return;

	// Handle port type.
	if (_net._port_index != 0 && (_net._iodirection == st_net::OUT ||
			_net._iodirection == st_net::INOUT))
		{
		// We only want to update the port if it's output or
		// inout direction.  We never update the port if it's
		// input or undefined.
		st_net *node = symboltable.get(*_net._port_index)->get_net();
		if (_net._port_ms >= 0 && _net._port_ls >= 0)
			node->assignment(*_net._storage, _net._port_ms, _net._port_ls);
		else
			node->assignment(*_net._storage);
		// Even if we are a port, we must still handle the local
		// variables.  So continue;
		}
	if (_net._propagate != 0)
		{
		st_net *node;
		list<hash_value>::iterator idx(_net._propagate->begin());
		list<hash_value>::iterator last(_net._propagate->end());
		for (; idx != last; ++idx)
			{
			node = symboltable.get(*idx)->get_net();
			// Only propagate to input/inout ports.
			if (node->_iodirection != st_net::IN &&
					node->_iodirection != st_net::INOUT)
				continue;
			node->assignment(*_net._storage);
			}
		}

	// Now check to see whether this variable was monitored.
	// If so, add the statements in the list to the event queue.
	if (_net._monitor != 0)
		check_monitor(*(_net._monitor), before, after);
	}


// Complete st_net check function object.

st_net_complete::st_net_complete(const st_net &n)
	: _net(n)
	{}

bool
st_net_complete::operator()() const
	{
	if ((_net._iodirection == st_net::IN || _net._iodirection == st_net::INOUT)
	 && _net._port_index != 0)
	 	{
		net_type *net = symboltable.get(*_net._port_index)->get_net();
		if (net != 0)
			return net->complete();
		}
	else
		return (_net._storage != 0);
	return false;
	}


// Evaluate st_net function object.

st_net_evaluate::st_net_evaluate(const st_net &n)
	: _net(n)
	{}

st_net_evaluate::num_type
st_net_evaluate::operator()(position_type m, position_type l) const
	{
	num_type value;
	if ((_net._iodirection == st_net::IN || _net._iodirection == st_net::INOUT)
	 && _net._port_index != 0)
		{
		net_type *net = symboltable.get(*_net._port_index)->get_net();
		st_net_evaluate eval(*net);
		value = eval(m, l);
		}
	else
		{
		if (m >= 0 && l < 0)
			value = (*(_net._storage))(m, m);
		else if (m >= 0 && l >= 0)
			value = (*(_net._storage))(m, l);
		else
			value = *(_net._storage);
		}
	return value;
	}
