// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// time_whl.h
//
// Time wheel class.  This class is a template class for the same reason
// as the "Events" class.  See Event.h for an explanation.

#ifndef _TIME_WHL_H
#define _TIME_WHL_H

#if defined(DEBUG_FUTURE)
#include <iostream.h> // cout debug statements, remove later...
#endif
#include <list>

class ostream;
template<class T> struct time_wheel_handler;

template<class T>
class time_wheel_node
	{
public:
	typedef unsigned long time_type;
	typedef T *data_type;
	typedef list<data_type> event_list_type;
	//typedef time_wheel_handler<T> handler_type;
	typedef ostream ostream_type;

	explicit time_wheel_node(time_type t = 0)
		: _time(t) {}
	time_wheel_node(const time_wheel_node<T> &tw)
		: _time(tw._time)
		{ append(tw._event); }

	void sim_add_event(T *op)
		{ _event.push_back(op); }
	time_type time() const
		{ return _time; }
	ostream_type &display(ostream_type &s) const
		{
		event_list_type::const_iterator idx(_event.begin());
		event_list_type::const_iterator last(_event.end());
		s << "Time " << _time << ":" << endl;
		for (; idx != last; ++idx)
			(*idx)->display(s);
		return s;
		}
	void sim_trigger(const time_wheel_handler<T> &);

protected:
	void append(const list<T *> &ev)
		{
		back_insert_iterator<event_list_type> bii(_event);
		copy(ev.begin(), ev.end(), bii);
		}

private:
	time_type _time; // Current time for this node. Absolute time.
	event_list_type _event; // List of events.
	};

template<class T>
struct time_wheel_handler
	{
	typedef time_wheel_node<T> node_type;
	typedef node_type::time_type time_type;
	typedef node_type::event_list_type event_list_type;

	// User function to perform operations on our object.
	virtual void operator()(time_type, event_list_type &) const
		{ cout << "No time wheel handler, using default!" << endl; }
	};

template<class T> inline void
time_wheel_node<T>::sim_trigger(const time_wheel_handler<T> &h)
	{ h(_time, _event); }


// Time Wheel.  We can not use a list to store the nodes in this case
// because we need to insert nodes.  Unless in the future we extend
// the list class to include sorted lists.


template<class T>
class time_wheel
	{
public:
	typedef time_wheel_node<T> node_type;
	typedef node_type::time_type time_type;
	//typedef node_type::handler_type handler_type;
	typedef node_type::ostream_type ostream_type;
	typedef list<node_type> node_list;

	time_wheel()
		// Create time zero, so we can test it later.
		{ _node.push_back(node_type()); }

	void sim_add_event(time_type t, T *obj)
		{
		// Find out where this object belongs.
		node_list::iterator idx(_node.begin());
		node_list::iterator stop(_node.end());
		for (; idx != stop; idx++)
			{
			if (t == (*idx).time())
				{
#if defined(DEBUG_FUTURE)
cout << "DEBUG_FUTURE:  appending to current time: " << (*idx).Time() << endl;
#endif
				(*idx).sim_add_event(obj);
				break;
				}
			else if (t > (*idx).time())
				{
				// Greater than current time unit.
				// Could insert if less than next time unit,
				// or create new node if next time unit doesn't
				// exist.
				node_list::iterator next(idx);
				++next;
				if (next != stop)
					{
					// Between this time unit and next time
					// unit, insert.
					if (t < (*next).time())
						{
#if defined(DEBUG_FUTURE)
cout << "DEBUG_FUTURE:  inserting time unit: " << (*next).Time() << endl;
#endif
						node_list::iterator tmp(
							_node.insert(next, node_type(t)));
						(*tmp).sim_add_event(obj);
						break;
						}
					// Greater or equal?  Go there and check!
					}
				else
					{
#if defined(DEBUG_FUTURE)
cout << "DEBUG_FUTURE:  appending to time unit: " << time << endl;
#endif
					// Hmm, at the end of our list, append.
					_node.push_back(node_type(t));
					node_list::iterator prev(_node.end());
					--prev;
					(*prev).sim_add_event(obj);
					break;
					}
				}
			else
				{
				cout << "Less than current time (Error)?" << endl << endl;
				return;
				}
			}
		}
	void sim_trigger(const time_wheel_handler<T> &h)
		{
		// Trigger our node.
		while (!_node.empty())
			{
			node_type &tmp(_node.front());
			tmp.sim_trigger(h);
			_node.pop_front();
			}
		}
	time_type current_time() const
		{ return _node.front().time(); }
	ostream_type &display(ostream_type &s) const
		{
		node_list::const_iterator idx(_node.begin());
		node_list::const_iterator last(_node.end());
		for (; idx != last; ++idx)
			(*idx).display(s);
		return s;
		}

private:
	node_list _node;
	};

#endif // _TIME_WHL_H
