// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// esetup.cc

#include <strstream.h>
#include "common/error.h"
#include "common/sym_tab.h"
#include "common/st_func.h"
#include "common/st_net.h"
#include "common/st_setup.h"
#include "common/st_util.h"
#include "expr/esetup.h"
#include "expr/eeval.h"
#include "expr/add.h"
#include "expr/andand.h"
#include "expr/eqeq.h"
#include "expr/funcenbl.h"
#include "expr/grteq.h"
#include "expr/grtthan.h"
#include "expr/invert.h"
#include "expr/leseq.h"
#include "expr/lesthan.h"
#include "expr/logand.h"
#include "expr/logor.h"
#include "expr/logxor.h"
#include "expr/lshift.h"
#include "expr/not.h"
#include "expr/noteq.h"
#include "expr/oror.h"
#include "expr/qstr.h"
#include "expr/rangeid.h"
#include "expr/rshift.h"
#include "expr/sub.h"
#include "misc/msetup.h"

extern symbol_table symboltable;

setup_expr::setup_expr(scopelist_type &s)
	: _scope(s)
	{}

void
setup_expr::operator()(add *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(and_and *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(equal_equal *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(func_enable *p) const
	{
	p->_index = symboltable.find(p->_name, _scope);
	if (p->_index._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM)p->_index._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}

	// Make sure it's a function.
	func_type *node = symboltable.get(p->_index)->get_function();
	if (node == 0)
		{
		sim_seterror(SE_NFUNCTION, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	// Functions are not setup until the function is called.  This is a
	// function call, so setup...
	node->sim_setup(setup_st_node(_scope, 0, p->_argument));

	// Now check argument list.
	if (!check_arguments(p->_argument, node->_iovariable))
		{
		sim_seterror(SE_NARGLST, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	if (p->_argument)
		{
		arg_list::iterator itp(p->_argument->begin());
		arg_list::iterator stop(p->_argument->end());
		for (; itp != stop; ++itp)
			(*itp)->sim_setup(setup_expr(_scope));
		}
	}

void
setup_expr::operator()(grt_equal *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(grt_than *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(invert *p) const
	{ p->_expr->sim_setup(setup_expr(_scope)); }

void
setup_expr::operator()(left_shift *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(les_equal *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(les_than *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(logic_and *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(logic_or *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(logic_xor *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(not_expr *p) const
	{ p->_expr->sim_setup(setup_expr(_scope)); }

void
setup_expr::operator()(not_equal *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(or_or *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(range_id *p) const
	{
	// First find the index.
	if (p->_index._scope < 0)
		{
		p->_index = symboltable.find(p->_name, _scope);
		if (p->_index._scope < 0)
			{
			sim_seterror((SIM_ERR_NUM) p->_index._value,
				p->_filename, p->_lineno);
			sim_perror(p->_name);
			}
		}

	num_type *tmp = 0;
	net_type *net = symboltable.get(p->_index)->get_net();
	// In the case of an lvalue in an assignment, a function name could
	// also be an lvalue.
	func_type *func = symboltable.get(p->_index)->get_function();
	if (net != 0)
		tmp = net->_storage;
	else if (func != 0)
		tmp = &(func->_storage);
	else
		{
		sim_seterror(SE_COMPLETE, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}

	// Check for valid range...
	range_id::position_type ms, ls;
	if (p->_select != 0)
		p->_select->sim_setup(setup_select(_scope));
	if (p->get_select(ms, ls) == range_id::select_type::SELECT_RANGE)
		{
		if ((*tmp)(ms,ls).size() == 0)
			{
			strstream buf;
			buf << *(p->_select);
			sim_seterror(SE_IDXORDER, p->_filename, p->_lineno);
			sim_perror(buf);
			}
		}
	}

void
setup_expr::operator()(right_shift *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}

void
setup_expr::operator()(sub *p) const
	{
	p->_left->sim_setup(setup_expr(_scope));
	p->_right->sim_setup(setup_expr(_scope));
	}
