// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// exprbase.h
//
// Base class for all expressions.

#ifndef _EXPRBASE_H
#define _EXPRBASE_H

#include "common/bvector.h"
#include "common/combase.h"

class add;
class and_and;
class equal_equal;
class func_enable;
class grt_equal;
class grt_than;
class invert;
class left_shift;
class les_equal;
class les_than;
class logic_and;
class logic_or;
class logic_xor;
class not_expr;
class not_equal;
class number;
class or_or;
class qouted_str;
class right_shift;
class range_id;
class sub;

struct expr_monitor;
struct expr_evaluate;
struct expr_setup;
struct expr_trigger;

struct expr_base : public virtual common_base
	{
	typedef bit_vector num_type;

	typedef expr_monitor monitor_type;
	typedef expr_evaluate evaluate_type;
	typedef expr_setup setup_type;
	typedef expr_trigger trigger_type;

	expr_base()
		{}
	expr_base(const expr_base &p)
		: common_base(p) {}
	virtual ~expr_base()
		{}

	// Functions which evaluate/setup/trigger functions need.
	virtual func_enable *get_func_enable()
		{ return 0; }
	virtual qouted_str *get_qouted_str()
		{ return 0; }
	virtual range_id *get_range_id()
		{ return 0; }

	virtual expr_base *copy_constructor() const
		{ return new expr_base(*this); }
	virtual ostream_type &display(ostream_type &s) const
		{ s << "(" << _filename << "/" << _lineno << ")"; return s; }

	virtual bool sim_monitor(const monitor_type &) const
		{ return false; } // Return success or failure.
	virtual num_type sim_evaluate(const evaluate_type &) const
		{ return num_type(DC); }
	virtual void sim_setup(const setup_type &)
		{}
	virtual void sim_trigger(const trigger_type &)
		{}
	};

inline ostream &
operator<<(ostream &s, const expr_base &d)
	{ return d.display(s); }

ostream &display_unary_expr(ostream &, const expr_base &, const char);
ostream &display_binary_expr(ostream &, const expr_base &, const expr_base &,
	const char *);

struct expr_monitor
	{
	virtual bool operator()(const add *) const
		{ return false; }
	virtual bool operator()(const and_and *) const
		{ return false; }
	virtual bool operator()(const equal_equal *) const
		{ return false; }
	virtual bool operator()(const func_enable *) const
		{ return false; }
	virtual bool operator()(const grt_equal *) const
		{ return false; }
	virtual bool operator()(const grt_than *) const
		{ return false; }
	virtual bool operator()(const invert *) const
		{ return false; }
	virtual bool operator()(const left_shift *) const
		{ return false; }
	virtual bool operator()(const les_equal *) const
		{ return false; }
	virtual bool operator()(const les_than *) const
		{ return false; }
	virtual bool operator()(const logic_and *) const
		{ return false; }
	virtual bool operator()(const logic_or *) const
		{ return false; }
	virtual bool operator()(const logic_xor *) const
		{ return false; }
	virtual bool operator()(const not_expr *) const
		{ return false; }
	virtual bool operator()(const not_equal *) const
		{ return false; }
	virtual bool operator()(const number *) const
		{ return false; }
	virtual bool operator()(const or_or *) const
		{ return false; }
	virtual bool operator()(const qouted_str *) const
		{ return false; }
	virtual bool operator()(const range_id *) const
		{ return false; }
	virtual bool operator()(const right_shift *) const
		{ return false; }
	virtual bool operator()(const sub *) const
		{ return false; }
	};

struct expr_evaluate
	{
	typedef expr_base::num_type num_type;

	virtual num_type operator()(const add *) const
		{ return num_type(DC); }
	virtual num_type operator()(const and_and *) const
		{ return num_type(DC); }
	virtual num_type operator()(const equal_equal *) const
		{ return num_type(DC); }
	virtual num_type operator()(const func_enable *) const
		{ return num_type(DC); }
	virtual num_type operator()(const grt_equal *) const
		{ return num_type(DC); }
	virtual num_type operator()(const grt_than *) const
		{ return num_type(DC); }
	virtual num_type operator()(const invert *) const
		{ return num_type(DC); }
	virtual num_type operator()(const left_shift *) const
		{ return num_type(DC); }
	virtual num_type operator()(const les_equal *) const
		{ return num_type(DC); }
	virtual num_type operator()(const les_than *) const
		{ return num_type(DC); }
	virtual num_type operator()(const logic_and *) const
		{ return num_type(DC); }
	virtual num_type operator()(const logic_or *) const
		{ return num_type(DC); }
	virtual num_type operator()(const logic_xor *) const
		{ return num_type(DC); }
	virtual num_type operator()(const not_expr *) const
		{ return num_type(DC); }
	virtual num_type operator()(const not_equal *) const
		{ return num_type(DC); }
	virtual num_type operator()(const number *) const
		{ return num_type(DC); }
	virtual num_type operator()(const or_or *) const
		{ return num_type(DC); }
	virtual num_type operator()(const qouted_str *) const
		{ return num_type(DC); }
	virtual num_type operator()(const range_id *) const
		{ return num_type(DC); }
	virtual num_type operator()(const right_shift *) const
		{ return num_type(DC); }
	virtual num_type operator()(const sub *) const
		{ return num_type(DC); }
	};

struct expr_setup
	{
	virtual void operator()(add *) const {}
	virtual void operator()(and_and *) const {}
	virtual void operator()(equal_equal *) const {}
	virtual void operator()(func_enable *) const {}
	virtual void operator()(grt_equal *) const {}
	virtual void operator()(grt_than *) const {}
	virtual void operator()(invert *) const {}
	virtual void operator()(left_shift *) const {}
	virtual void operator()(les_equal *) const {}
	virtual void operator()(les_than *) const {}
	virtual void operator()(logic_and *) const {}
	virtual void operator()(logic_or *) const {}
	virtual void operator()(logic_xor *) const {}
	virtual void operator()(not_expr *) const {}
	virtual void operator()(not_equal *) const {}
	virtual void operator()(number *) const {}
	virtual void operator()(or_or *) const {}
	virtual void operator()(qouted_str *) const {}
	virtual void operator()(range_id *) const {}
	virtual void operator()(right_shift *) const {}
	virtual void operator()(sub *) const {}
	};

struct expr_trigger
	{
	virtual void operator()(add *) const {}
	virtual void operator()(and_and *) const {}
	virtual void operator()(equal_equal *) const {}
	virtual void operator()(func_enable *) const {}
	virtual void operator()(grt_equal *) const {}
	virtual void operator()(grt_than *) const {}
	virtual void operator()(invert *) const {}
	virtual void operator()(left_shift *) const {}
	virtual void operator()(les_equal *) const {}
	virtual void operator()(les_than *) const {}
	virtual void operator()(logic_and *) const {}
	virtual void operator()(logic_or *) const {}
	virtual void operator()(logic_xor *) const {}
	virtual void operator()(not_expr *) const {}
	virtual void operator()(not_equal *) const {}
	virtual void operator()(number *) const {}
	virtual void operator()(or_or *) const {}
	virtual void operator()(qouted_str *) const {}
	virtual void operator()(range_id *) const {}
	virtual void operator()(right_shift *) const {}
	virtual void operator()(sub *) const {}
	};

#endif // _EXPRBASE_H
