// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// dectrig.cc

#include "common/error.h"
#include "common/time_whl.h"
#include "common/sym_tab.h"
#include "common/st_net.h"
#include "common/st_eval.h"
#include "expr/eeval.h"
#include "expr/emon.h"
#include "misc/delaynum.h"
#include "misc/delayid.h"
#include "misc/evntexpr.h"
#include "misc/dectrig.h"

extern symbol_table symboltable;
extern time_wheel<stmt_base> timewheel;


// Event expression trigger object.

trigger_event_expr::trigger_event_expr(event_cache_type *e)
	{ _cache = e; }

void
trigger_event_expr::operator()(event_expr *p) const
	{
	num_type initial_value = p->_expr->sim_evaluate(evaluate_expr());
	event_type *evnt = 0;
	if (p->_type == TRIG_BITCHNG)
		evnt = new bit_change_event<stmt_type>(_cache, initial_value);
	else if (p->_type == TRIG_POSEDGE)
		evnt = new posedge_event<stmt_type>(_cache, initial_value);
	else if (p->_type == TRIG_NEGEDGE)
		evnt = new negedge_event<stmt_type>(_cache, initial_value);

	monitor_expr::event_ptr eptr(evnt);
	// Setup should have checked for validity.
	p->_expr->sim_monitor(monitor_expr(&eptr));
	evnt->monitor_bit(p->_expr);
	// The event object now owns this expression.
	p->_expr = 0;
	}


// Delay or event control trigger object.

trigger_dec::trigger_dec(stmt_type *st)
	{ _stmt = st; }

void
trigger_dec::operator()(delay_num *p) const
	{
	// Find the current time and add the delay.
	// Append to that time unit.
	time_wheel<stmt_base>::time_type t = timewheel.current_time();
	t += (unsigned long) *p;
	timewheel.sim_add_event(t, _stmt);
	}

void
trigger_dec::operator()(delay_id *p) const
	{
	// The delay is specified by the identifier.
	net_type *node = symboltable.get(p->_index)->get_net();
	time_wheel<stmt_base>::time_type t = timewheel.current_time();
	num_type amt(node->sim_evaluate(evaluate_st_node()));
	t += (unsigned long) amt;
	timewheel.sim_add_event(t, _stmt);
	}

void
trigger_dec::operator()(ored_event_expr *p) const
	{
	// Only one Event object can exist for the entire ored event list.
	// Otherwise, we will trigger two events when we should only
	// trigger one.  Add this event to the list for all events in
	// this expression.
	event_cache_type *cache = new event_cache_type(_stmt);
	// If _parent is part of an initial procedural block, mark the
	// cache so that it is only triggered once.
	if (_stmt->_always == false)
		cache->_event_state = event_cache_type::ES_VOLATILE;
	ored_event_expr::event_expr_list::iterator itp(p->_ored_list->begin());
	ored_event_expr::event_expr_list::iterator stop(p->_ored_list->end());
	for (; itp != stop; ++itp)
		(*itp).get()->sim_trigger(trigger_event_expr(cache));

	// Once an ored event expression is triggered, it is no longer
	// needed.
	if (_stmt->_dec != 0)
		{
		delete _stmt->_dec;
		_stmt->_dec = 0;
		}
	}
