// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// d_setup.cc

#include "common/error.h"
#include "common/sym_tab.h"
#include "expr/rangeid.h"
#include "moditm/d_reg.h"
#include "moditm/d_wire.h"
#include "moditm/d_in.h"
#include "moditm/d_out.h"
#include "moditm/d_inout.h"
#include "moditm/d_setup.h"

extern symbol_table symboltable;

decl_setup::decl_setup(scopelist_type &s, io_list *i)
	: _scope(s) { _iovariable = i; }

void
decl_setup::net_setup(ident_list *il, range_type *r, type t) const
	{
	// Go through each variable, and set it up.
	ident_list::iterator itp(il->begin());
	ident_list::iterator stop(il->end());
	for (; itp != stop; ++itp)
		{
		// Find the symbol table node, if it exists.
		hash_type hv(symboltable.find((*itp)->name(), _scope));
		net_type *net = 0;
		if (hv._scope < 0)
			{
			// Does not exist, create one.
			net = new net_type((*itp)->name(), t);
			net->_scope = _scope.top();
			hv = symboltable.add(net);
			}
		else
			net = symboltable.get(hv)->get_net();
		// Allocate storage space, make sure this is the first time.
		if (net->_storage == 0)
			{
			range_type::position_type ms = 0, ls = 0;
			if (r != 0)
				r->get_data(ms, ls);
			net->_storage = new num_type(ms, ls);
			}
		// net->_storage != 0
		else if (t == net_type::WIRE)
			{
			sim_seterror(SE_WIREDECL, (*itp)->_filename, (*itp)->_lineno);
			sim_perror((*itp)->name());
			}
		// t == net_type::REGISTER
		else if (net->_type == net_type::REGISTER
			|| net->_iodirection == net_type::IN
			|| net->_iodirection == net_type::INOUT)
			{
			sim_seterror(SE_REGDECL, (*itp)->_filename, (*itp)->_lineno);
			sim_perror((*itp)->name());
			}
		else
			net->_type = net_type::REGISTER;

		// Update filename and line number.
		net->_filename = (*itp)->_filename;
		net->_lineno = (*itp)->_lineno;
		}
	}

void
decl_setup::operator()(reg_decl *p) const
	{ net_setup(p->_id_list, p->_range, net_type::REGISTER); }

void
decl_setup::operator()(wire_decl *p) const
	{ net_setup(p->_id_list, p->_range, net_type::WIRE); }

void
decl_setup::io_setup(ident_list *il, range_type *r, iodirection_type iod) const
	{
	// Setup a local scope stack for the local variable setup.
	scopelist_type localscope;
	localscope.push(_scope.top());

	// Now go through the list of ids to be declared.
	ident_list::iterator itp(il->begin());
	ident_list::iterator stop(il->end());
	for (; itp != stop; ++itp)
		{
		hash_type hv(symboltable.find((*itp)->name(), localscope));
		net_type *net;
		if (hv._scope < 0)
			{
			// Does not exist, create one.
			net = new net_type((*itp)->name(), net_type::WIRE);
			net->_scope = localscope.top();
			hv = symboltable.add(net);
			}
		else
			{
			net = symboltable.get(hv)->get_net();
			if (net == 0)
				{
				sim_seterror(SE_TYPE, (*itp)->_filename, (*itp)->_lineno);
				sim_perror((*itp)->name());
				}
			}

		// Check the net object for validity.
		if (net->_iodirection != net_type::UNDEF)
			{
			sim_seterror(SE_IODECL, (*itp)->_filename, (*itp)->_lineno);
			sim_perror((*itp)->name());
			}
		else
			net->_iodirection = iod;

		// Input and inout variables need storage for cascading data.
		if (net->_iodirection == net_type::IN
		 || net->_iodirection == net_type::INOUT)
			{
			if (net->_storage == 0)
				{
				range_type::position_type ms = 0, ls = 0;
				if (r != 0)
					r->get_data(ms, ls);
				net->_storage = new num_type(ms, ls);
				}
			}

		// Save the hash value index for later testing.
		if (_iovariable != 0)
			_iovariable->push_back(hv);

		// Update filename and line number.
		net->_filename = (*itp)->_filename;
		net->_lineno = (*itp)->_lineno;
		}
	}

void
decl_setup::operator()(input_decl *p) const
	{ io_setup(p->_id_list, p->_range, net_type::IN); }

void
decl_setup::operator()(output_decl *p) const
	{ io_setup(p->_id_list, p->_range, net_type::OUT); }

void
decl_setup::operator()(inout_decl *p) const
	{ io_setup(p->_id_list, p->_range, net_type::INOUT); }
