// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// misetup.cc

#include "common/error.h"
#include "common/sym_tab.h"
#include "common/time_whl.h"
#include "common/st_func.h"
#include "common/st_inst.h"
#include "common/st_task.h"
#include "common/st_mod.h"
#include "common/st_setup.h"
#include "expr/esetup.h"
#include "stmt/ssetup.h"
#include "misc/range.h" // Function return value is a range.
#include "misc/dectrig.h"
#include "misc/msetup.h"
#include "moditm/initial.h"
#include "moditm/always.h"
#include "moditm/cassign.h"
#include "moditm/func.h"
#include "moditm/task.h"
#include "moditm/modinst.h"
#include "moditm/misetup.h"

extern time_wheel<stmt_base> timewheel;
extern symbol_table symboltable;

struct setup_module_instance : public module_instance_setup
	{
	typedef Stack<int> scopelist_type;
	typedef st_node_base node_type;
	typedef node_type::hash_type hash_type;
	typedef module_instance::port_conn_list port_conn_list;
	typedef st_instantiation instantiation_type;
	typedef st_module module_type;

	setup_module_instance(scopelist_type &, module_type *);
	void operator()(module_instance *) const;
	scopelist_type &_scope;
	module_type *_stmod;
	};

setup_module_instance::setup_module_instance(scopelist_type &s, module_type *m)
	: _scope(s)
	{ _stmod = m; }

void
setup_module_instance::operator()(module_instance *p) const
	{
	// Create a module instance symbol table object.
	hash_type hv(symboltable.find(p->_name, _scope));
	if (hv._scope < 0)
		{
		node_type *inst = new instantiation_type(p->_name);
		inst->_filename =p->_filename;
		inst->_lineno = p->_lineno;
		symboltable.add(inst);
		}
	else
		{
		sim_seterror(SE_STDUP, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}

	// Setup the port connections.
	port_conn_list::iterator itp(p->_port_conn->begin());
	port_conn_list::iterator stop(p->_port_conn->end());
	for (; itp != stop; ++itp)
		(*itp)->sim_setup(setup_port_conn(_scope));

	// Setup the module definition being used.
	_stmod->sim_setup(setup_st_node(_scope, p->_port_conn));
	}


setup_module_item::setup_module_item(scopelist_type &s, hash_list *i)
	: decl_setup(s, i) {}

void
setup_module_item::operator()(initial *p) const
	{
	// First, setup the statement, but don't modify the original.
	initial::stmt_type *st = p->_stmt->copy_constructor();
	st->sim_setup(setup_stmt(_scope));
	// If no delay or event control, put statement in timewheel, at
	// time zero.  DEC should already be setup by the statement setup.
	if (st->_dec != 0)
		{
		st->_dec->sim_trigger(trigger_dec(st));
		delete st->_dec;
		st->_dec = 0;
		}
	else
		timewheel.sim_add_event(0, st);
#ifdef DEBUG_INIT
	cout << "INIT:  initial statement appended to time wheel." << endl;
#endif
	}

void
setup_module_item::operator()(always *p) const
	{
	// Some initial setup.
	always::stmt_type *st = p->_stmt->copy_constructor();
	st->_always = true;
	st->sim_setup(setup_stmt(_scope));
	// DEC should be none-null, but check anyway.  DEC should already
	// be setup by the statement setup.
	if (st->_dec != 0)
		st->_dec->sim_trigger(trigger_dec(st));
	else
		{
		sim_seterror(SE_SUPPORT, p->_filename, p->_lineno);
		sim_perror("always(delay or event missing)");
		}
	}

void
setup_module_item::operator()(cont_assign *p) const
	{
	// A continuous assignment has many similarities with other
	// Verilog constructs.  One of them is the 'always' procedural
	// block.  Since a change in any one of the variables in the
	// right-hand-side of the assignment will cause an event to occur.
	// It also has a delay control.  Thus like a loop, we can not
	// delete the control.  So we must set it up correctly.
	if (p->_dec != 0)
		p->_dec->_persist = true;
	cont_assign::assign_stmt_list::iterator itp(p->_assign->begin());
	cont_assign::assign_stmt_list::iterator stop(p->_assign->end());
	event_cache_type *cache = 0;
	event_type *ev = 0;
	cont_assign::assign_stmt *st = 0;
	for (; itp != stop; ++itp)
		{
		st = (*itp)->copy_constructor();
		// This delay or event control should only be a delay literal,
		// i.e. delay_num.  No need setup.
		if (p->_dec != 0)
			st->_dec = p->_dec->copy_constructor();
		cache = new event_cache_type(st);
		ev = new change_event<stmt_type>(cache, DC);
		counted_ptr< event_base<stmt_type> > e(ev);
		st->sim_setup(setup_stmt(_scope, &e));
		}
	}

void
setup_module_item::operator()(function *p) const
	{
	// Setup function definitions mean saving it in the symbol table.
	hash_type hv(symboltable.find(p->_name, _scope));
	func_type *node;
	if (hv._scope < 0)
		{
		// Does not exist, create a new symbol table node.
		// Make a copy of the statement, because each new function
		// definition could have a different scope if the module it
		// is in is instantiated more than once.
		function::stmt_type *st = p->_stmt->copy_constructor();
		node = new func_type(p->_name, st, p->_decl);
		node->_scope = _scope.top();
		function::range_type::position_type ms = 0, ls = 0;
		if (p->_retsize != 0)
			p->_retsize->get_data(ms, ls);
		node->_storage = num_type(ms, ls);
		node->_filename = p->_filename;
		node->_lineno = p->_lineno;
		symboltable.add(node);
		}
	else
		{
		sim_seterror(SE_STDUP, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	}

void
setup_module_item::operator()(task *p) const
	{
	// Setup task definitions mean saving it in the symbol table.
	hash_type hv(symboltable.find(p->_name, _scope));
	task_type *node;
	if (hv._scope < 0)
		{
		// Does not exist, create a new symbol table node.
		// Make a copy of the statement, because each new function
		// definition could have a different scope if the module it
		// is in is instantiated more than once.
		task::stmt_type *st = p->_stmt->copy_constructor();
		node = new task_type(p->_name, st, p->_decl);
		node->_scope = _scope.top();
		node->_filename = p->_filename;
		node->_lineno = p->_lineno;
		symboltable.add(node);
		}
	else
		{
		sim_seterror(SE_STDUP, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	}

void
setup_module_item::operator()(module_instantiation *p) const
	{
	// Find the module definition for setting up the instances.
	hash_type hv(symboltable.find(p->_name, _scope));
	if (hv._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM) hv._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	module_type *mod = symboltable.get(hv)->get_module();
	if (mod == 0)
		{
		sim_seterror(SE_TYPE, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}

	// Setup each instance of the instantiation.
	module_instantiation::mod_inst_list::iterator itp(p->_instance->begin());
	module_instantiation::mod_inst_list::iterator stop(p->_instance->end());
	for (; itp != stop; ++itp)
		(*itp)->sim_setup(setup_module_instance(_scope, mod));
	}


// Front ends to make virtual calls work.


void
setup_module_item::operator()(reg_decl *p) const
	{ decl_setup::operator()(p); }

void
setup_module_item::operator()(wire_decl *p) const
	{ decl_setup::operator()(p); }

void
setup_module_item::operator()(input_decl *p) const
	{ decl_setup::operator()(p); }

void
setup_module_item::operator()(output_decl *p) const
	{ decl_setup::operator()(p); }

void
setup_module_item::operator()(inout_decl *p) const
	{ decl_setup::operator()(p); }
