// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// p_expr.cc

#include <stdlib.h>
#include <string>
#include <list>
#include "common/bvector.h"
#include "expr/add.h"
#include "expr/andand.h"
#include "expr/eqeq.h"
#include "expr/funcenbl.h"
#include "expr/invert.h"
#include "expr/grtthan.h"
#include "expr/grteq.h"
#include "expr/lesthan.h"
#include "expr/leseq.h"
#include "expr/logand.h"
#include "expr/logor.h"
#include "expr/logxor.h"
#include "expr/lshift.h"
#include "expr/not.h"
#include "expr/noteq.h"
#include "expr/number.h"
#include "expr/oror.h"
#include "expr/qstr.h"
#include "expr/rangeid.h"
#include "expr/rshift.h"
#include "expr/sub.h"
#include "parser.h"
#include "p_expr.h"

p_expressionlist
p_create_expressionlist(p_expressionlist lst, p_expression exp)
	{
	list< basic_ptr<expr_base> > *ret;
	if (lst != 0)
		ret = (list< basic_ptr<expr_base> > *) lst;
	else
		ret = new list< basic_ptr<expr_base> >;
	if (exp != 0)
		{
		ret->push_back((expr_base *) exp);
#ifdef DEBUG_PARSER
		cout << "PARSER:  added expression to list." << endl;
#endif
		}
	return ret;
	}

p_rangeidlist
p_create_rangeidlist(p_rangeidlist idlst, p_rangeid id)
	{
	list< basic_ptr<range_id> > *ret;
	if (idlst != 0)
		ret = (list< basic_ptr<range_id> > *)idlst;
	else
		ret = new list< basic_ptr<range_id> >;
	if (id != 0)
		{
		ret->push_back((range_id *) id);
#ifdef DEBUG_PARSER
		cout << "PARSER:  added range id to list." << endl;
#endif
		}
	return ret;
	}

p_rangeid
p_create_rangeid(char *n, p_select r)
	{
	range_id *ret;
	ret = new range_id(n, (select_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
	free(n);
#ifdef DEBUG_PARSER
	cout << "PARSER:  created range id." << endl;
#endif
	return ret;
	}

p_qstring
p_create_qstring(char *str)
	{
	qouted_str *ret;
	ret = new qouted_str(str);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
	free(str);
#ifdef DEBUG_PARSER
	cout << "PARSER:  created qouted string." << endl;
#endif
	return ret;
	}

p_number
p_create_number(char *str)
	{
	bit_vector::size_type len = 0;
	bit_vector::base_type type = bit_vector::BASE2;
	char *siz, *num, base;
	number *ret;
	if (strchr(str, '\'') == 0)
		{
		base = 'd';
		num = str;
		}
	else
		{
		char *tmp = my_strdup(str);
		siz = strtok(tmp, "\'");
		len = atoi(siz);
		base = *(str + strlen(siz) + 1);
		num = str + strlen(siz) + 2;
		free(tmp);
		}
	switch(base)
		{
		case 'b': case 'B': type = bit_vector::BASE2; break;
		case 'o': case 'O': type = bit_vector::BASE8; break;
		case 'd': case 'D': type = bit_vector::BASE10; break;
		case 'h': case 'H': type = bit_vector::BASE16; break;
		}
	ret = new number(num, type, len);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
	free(str);
#ifdef DEBUG_PARSER
	cout << "PARSER:  created number." << endl;
#endif
	return ret;
	}

p_add
p_create_add(p_expression l, p_expression r)
	{
	add *ret;
	ret = new add((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created add expression." << endl;
#endif
	return ret;
	}

p_subtract
p_create_subtract(p_expression l, p_expression r)
	{
	sub *ret;
	ret = new sub((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created substract expression." << endl;
#endif
	return ret;
	}

p_invert
p_create_invert(p_expression exp)
	{
	invert *ret;
	ret = new invert((expr_base *) exp);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created ~ expression." << endl;
#endif
	return ret;
	}

p_logand
p_create_logand(p_expression l, p_expression r)
	{
	logic_and *ret;
	ret = new logic_and((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created & expression." << endl;
#endif
	return ret;
	}

p_logor
p_create_logor(p_expression l, p_expression r)
	{
	logic_or *ret;
	ret = new logic_or((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created | expression." << endl;
#endif
	return ret;
	}

p_logxor
p_create_logxor(p_expression l, p_expression r)
	{
	logic_xor *ret;
	ret = new logic_xor((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created ^ expression." << endl;
#endif
	return ret;
	}

p_lshift
p_create_lshift(p_expression l, p_expression r)
	{
	left_shift *ret;
	ret = new left_shift((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created left shift expression." << endl;
#endif
	return ret;
	}

p_rshift
p_create_rshift(p_expression l, p_expression r)
	{
	right_shift *ret;
	ret = new right_shift((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created right shift expression." << endl;
#endif
	return ret;
	}

p_not
p_create_not(p_expression exp)
	{
	not_expr *ret;
	ret = new not_expr((expr_base *) exp);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created ! expression." << endl;
#endif
	return ret;
	}

p_eqeq
p_create_eqeq(p_expression l, p_expression r)
	{
	equal_equal *ret;
	ret = new equal_equal((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created == expression." << endl;
#endif
	return ret;
	}

p_noteq
p_create_noteq(p_expression l, p_expression r)
	{
	not_equal *ret;
	ret = new not_equal((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created != expression." << endl;
#endif
	return ret;
	}

p_lesthan
p_create_lesthan(p_expression l, p_expression r)
	{
	les_than *ret;
	ret = new les_than((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created < expression." << endl;
#endif
	return ret;
	}

p_leseq
p_create_leseq(p_expression l, p_expression r)
	{
	les_equal *ret;
	ret = new les_equal((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created <= expression." << endl;
#endif
	return ret;
	}

p_grtthan
p_create_grtthan(p_expression l, p_expression r)
	{
	grt_than *ret;
	ret = new grt_than((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created > expression." << endl;
#endif
	return ret;
	}

p_grteq
p_create_grteq(p_expression l, p_expression r)
	{
	grt_equal *ret;
	ret = new grt_equal((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created >= expression." << endl;
#endif
	return ret;
	}

p_oror
p_create_oror(p_expression l, p_expression r)
	{
	or_or *ret;
	ret = new or_or((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created || expression." << endl;
#endif
	return ret;
	}

p_andand
p_create_andand(p_expression l, p_expression r)
	{
	and_and *ret;
	ret = new and_and((expr_base *) l, (expr_base *) r);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created && expression." << endl;
#endif
	return ret;
	}

p_funcenable
p_create_funcenable(char *name, p_expressionlist exp)
	{
	func_enable *ret;
	ret = new func_enable(name, (list< basic_ptr<expr_base> > *) exp);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created function call." << endl;
#endif
	return ret;
	}
