// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// p_stmt.cc

#include <list>
#include "common/error.h"
#include "expr/exprbase.h"
#include "stmt/seqblk.h"
#include "stmt/taskenbl.h"
#include "stmt/bassign.h"
#include "stmt/nbassign.h"
#include "stmt/if.h"
#include "stmt/ifelse.h"
#include "stmt/for.h"
#include "stmt/forever.h"
#include "stmt/repeat.h"
#include "stmt/while.h"
#include "stmt/case.h"
#include "parser.h"
#include "p_expr.h"
#include "p_stmt.h"
#include "p_misc.h"

p_lvalue
p_create_lvalue(p_rangeid rid)
	{
	lvalue *ret;
	ret = new lvalue((range_id *) rid);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created lvalue." << endl;
#endif
	return ret;
	}

p_caseitemlist
p_create_caseitemlist(p_caseitemlist cilst, p_caseitem ci)
	{
	list< basic_ptr<case_item> > *ret;
	if (cilst != 0)
		ret = (list< basic_ptr<case_item> > *)cilst;
	else
		ret = new list< basic_ptr<case_item> >;
	ret->push_back((case_item *)ci);
#ifdef DEBUG_PARSER
	cout << "PARSER:  added case item to list." << endl;
#endif
	return ret;
	}

p_caseitem
p_create_caseitem(p_expressionlist elst, p_stmts st)
	{
	case_item *ret;
	if (elst == 0)
		ret = new case_item((stmt_base *)st);
	else
		ret = new case_item((stmt_base *)st,
			(list< basic_ptr<expr_base> > *)elst);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created case item." << endl;
#endif
	return ret;
	}

p_stmtslist
p_create_stmtslist(p_stmtslist slst, p_stmts st)
	{
	list< basic_ptr<stmt_base> > *ret;
	if (slst != 0)
		ret = (list< basic_ptr<stmt_base> > *)slst;
	else
		ret = new list< basic_ptr<stmt_base> >;
	if (st != 0)
		{
		ret->push_back((stmt_base *) st);
#ifdef DEBUG_PARSER
		cout << "PARSER:  added statement to list." << endl;
#endif
		}
	return ret;
	}

p_stmts
p_create_stmts(void)
	{
	stmt_base *ret;
	ret = new null_stmt;
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created null statement." << endl;
#endif
	return ret;
	}

p_sequentialblock
p_create_sequentialblock(p_stmtslist slst)
	{
	seq_block_stmt *ret;
	ret = new seq_block_stmt((list< basic_ptr<stmt_base> > *)slst);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created sequential block." << endl;
#endif
	return ret;
	}

p_taskstmt
p_create_taskstmt(char *name, p_expressionlist explst)
	{
	task_enable_stmt *ret;
	ret = new task_enable_stmt(name, (list< basic_ptr<expr_base> > *) explst);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
	free(name);
#ifdef DEBUG_PARSER
	cout << "PARSER:  created task enable statement." << endl;
#endif
	return ret;
	}

p_blkassignstmt
p_create_blkassignstmt(p_lvalue lv, p_expression exp)
	{
	block_assign_stmt *ret;
	ret = new block_assign_stmt((lvalue *)lv, (expr_base *)exp);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created blocking assignment statement." << endl;
#endif
	return ret;
	}

p_nblkassignstmt
p_create_nblkassignstmt(p_lvalue lv, p_expression exp)
	{
	non_block_assign_stmt *ret;
	ret = new non_block_assign_stmt((lvalue *)lv, (expr_base *)exp);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created non-blocking assignment statement." << endl;
#endif
	return ret;
	}

p_ifstmt
p_create_ifstmt(p_expression exp, p_stmts st)
	{
	if_stmt *ret;
	ret = new if_stmt((expr_base *)exp, (stmt_base *)st);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created if statement." << endl;
#endif
	return ret;
	}

p_ifelsestmt
p_create_ifelsestmt(p_expression exp, p_stmts t, p_stmts f)
	{
	if_else_stmt *ret;
	ret = new if_else_stmt((expr_base *)exp, (stmt_base *)t, (stmt_base *)f);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created if/else statement." << endl;
#endif
	return ret;
	}

p_forstmt
p_create_forstmt(p_blkassignstmt pre, p_expression exp,
		p_blkassignstmt post, p_stmts st)
	{
	for_stmt *ret;
	ret = new for_stmt((block_assign_stmt *)pre, (expr_base *)exp,
		(block_assign_stmt *)post, (stmt_base *)st);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created for loop statement." << endl;
#endif
	return ret;
	}

p_foreverstmt
p_create_foreverstmt(p_stmts st)
	{
	forever_stmt *ret;
	ret = new forever_stmt((stmt_base *)st);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created forever loop statement." << endl;
#endif
	return ret;
	}

p_repeatstmt
p_create_repeatstmt(p_expression exp, p_stmts st)
	{
	repeat_stmt *ret;
	ret = new repeat_stmt((expr_base *)exp, (stmt_base *)st);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created repeat loop statement." << endl;
#endif
	return ret;
	}

p_whilestmt
p_create_whilestmt(p_expression exp, p_stmts st)
	{
	while_stmt *ret;
	ret = new while_stmt((expr_base *)exp, (stmt_base *)st);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created while loop statement." << endl;
#endif
	return ret;
	}

p_casestmt
p_create_casestmt(p_expression exp, p_caseitemlist cilst)
	{
	case_stmt *ret;
	ret = new case_stmt((expr_base *)exp,
		(list< basic_ptr<case_item> > *)cilst);
	ret->_filename = cur_filename;
	ret->_lineno = cur_lineno;
#ifdef DEBUG_PARSER
	cout << "PARSER:  created case statement." << endl;
#endif
	return ret;
	}

p_stmts
p_create_includedec(p_stmts st, p_delayevntctl dec)
	{
	stmt_base *ret;
	ret = (stmt_base *)st;
	if (ret->_dec != 0)
		{
		strstream buf;
		buf << *(dec_base *)dec;
		sim_seterror(SE_SUPPORT, cur_filename, cur_lineno);
		sim_perror(buf);
		}
	ret->_dec = (dec_base *)dec;
#ifdef DEBUG_PARSER
	cout << "PARSER:  added delay or event control to statement." << endl;
#endif
	return ret;
	}
