// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// sim.cc

#include <list>
#include "common/ptr.h"
#include "common/dstack.h"
#include "common/error.h"
#include "common/event.h"
#include "common/time_whl.h"
#include "common/sym_tab.h"
#include "common/st_mod.h"
#include "common/st_setup.h"
#include "misc/module.h"
#include "misc/port.h"
#include "misc/msetup.h"
#include "misc/dectrig.h"
#include "stmt/stmtbase.h"
#include "stmt/finish.h"
#include "stmt/write.h"
#include "stmt/monitor.h"
#include "stmt/time.h"
#include "stmt/ssetup.h"
#include "stmt/strigger.h"
#include "p_misc.h"
#include "parser.h"

time_wheel<stmt_base> timewheel;
event_queue<stmt_base> eventqueue;
symbol_table symboltable;

#ifdef __cplusplus
extern "C" {
#endif

void
sim_init(char *pn)
	{
	// Setup variables for error reporting.
	sim_program_name = pn;
	sim_lineno = -1;
	sim_errno = SE_NONE;

	// Symbol table for simulation
	symboltable.initialize(HASHSIZE);

	// Our global system tasks and functions.
	// System tasks and functions are always global, so we do not really
	// need to pass in any scope values.  But an argument is still needed.
	Stack<int> scope;
	systask_finish *stf = new systask_finish;
	stf->sim_setup(setup_stmt(scope));
	systask_write *stw = new systask_write;
	stw->sim_setup(setup_stmt(scope));
	systask_monitor *stm = new systask_monitor;
	stm->sim_setup(setup_stmt(scope));
	sysfunc_time *sft = new sysfunc_time;
	sft->sim_setup(setup_stmt(scope));
	}

void
sim_parsed_module(p_module m)
	{
	// After we parsed a complete module, we need to store it into
	// the symbol table for later use.  Currently, we limit the
	// name of the top-level module to "main".  This is because we
	// do not know how to find the top-level module in the symbol
	// table.
	module *mod = (module *)m;

#if defined(DEBUG_MODULE)
	cout << "Finished parsing a module:" << endl << *mod << endl;
#endif

	// Now enter the module into the symbol table.  The scope should
	// always be global.  There are no nested module definitions.
	st_module *stmod = new st_module(mod->name());
	stmod->scope(0);

	// Save port list and module items if any.
	mod->sim_setup(setup_module(stmod));

	// Enter node into symbol table.
	hash_value hv(symboltable.add(stmod));
	if (hv._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM) hv._value, mod->_filename, mod->_lineno);
		sim_perror(mod->name());
		}
	delete mod;
	}

// Function object to handle events in the event queue.
template<class T>
struct eqhandler : public event_queue_handler<T>
	{
	void operator()(T *st) const
		{ st->sim_trigger(trigger_stmt(st)); }
	};

// Function object to handle events in the time wheel.
template<class T>
struct twhandler : public time_wheel_handler<T>
	{
	void operator()(
#if defined(DEBUG_TW)
			time_wheel_handler<T>::time_type t,
#else
			time_wheel_handler<T>::time_type,
#endif
			event_list_type &tlst) const
		{
#if defined(DEBUG_TW)
cout << "DEBUG_TW:  Starting event time " << t << "..." << endl << endl;
#endif
		// Trigger the operations in this time event.
		while (!tlst.empty())
			{
			// Statements return 'true' for two reasons;  1.
			// It has finished exec'ing or 2.  It is finished
			// with one nesting level.  Thus, we can't just
			// delete this statement without knowing which
			// case it was.
			T *obj = tlst.front();
#if defined(DEBUG_TW)
cout << "DEBUG_TW:  Executing next event..." << endl;
#endif
			obj->sim_trigger(trigger_stmt(obj));
			tlst.pop_front();
			}
#if defined(DEBUG_TW)
cout << "DEBUG_TW:  Handling event queue..." << endl;
#endif
		eventqueue.sim_trigger(eqhandler<T>());
#if defined(DEBUG_TW)
cout << endl;
#endif
		}
	};

void
sim_start()
	{
	// First, parse all files.
	sim_state = SS_COMPILE;
	p_start(); 

	// Start work.  First we need to setup for simulation.  This means
	// grabbing the top-level module from the symbol table and call
	// sim_setup for this module.
	hash_value idx(hash("main", 0));
	st_module *toplevel = symboltable.get(idx)->get_module();
	if (toplevel == 0)
		{
		sim_seterror(SE_NTOPLVL, "internal", -1);
		sim_perror("main");
		}
	Stack<int> scope;
	scope.push(0);

#if defined(DEBUG_SYMTAB)
	cout << "DEBUG_SYMTAB:  Setting up for simulation..." << endl;
#endif

	sim_state = SS_SETUP;
	toplevel->sim_setup(setup_st_node(scope));

#if defined(DEBUG_SYMTAB)
	cout << "DEBUG_SYMTAB: Symbol Table:" << endl << symboltable
		<< endl << endl;
#endif
	cout << "Begin simulation:" << endl << endl;

	sim_state = SS_SIMULATE;
	timewheel.sim_trigger(twhandler<stmt_base>());

	cout << endl << "No more events, ending simulation." << endl << endl;
	}

void
sim_perror(char *type, char *mesg, char *word, char *fn, int ln)
	{
	if (strcmp(type, "symbol") == 0)
		sim_errno = SE_PSYMBOL;
	else if (strcmp(type, "keyword") == 0)
		sim_errno = SE_PKEYWORD;
	else
		sim_errno = SE_PEOF;
	sim_file_name = fn;
	sim_lineno = ln;
	strstream buf;
	buf << mesg << ": " << word;
	sim_perror(buf);
	}

void
sim_debug(char *)
	{ cout << symboltable << endl; }

#ifdef __cplusplus
}
#endif // __cplusplus
