// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// ssetup.cc

#include <strstream.h>
#include "common/error.h"
#include "common/sym_tab.h"
#include "common/st_func.h"
#include "common/st_task.h"
#include "common/st_setup.h"
#include "common/st_util.h"
#include "expr/qstr.h"
#include "expr/rangeid.h"
#include "expr/esetup.h"
#include "expr/emon.h"
#include "misc/decsetup.h"
#include "stmt/lvalue.h"
#include "stmt/finish.h"
#include "stmt/write.h"
#include "stmt/monitor.h"
#include "stmt/time.h"
#include "stmt/seqblk.h"
#include "stmt/taskenbl.h"
#include "stmt/bassign.h"
#include "stmt/nbassign.h"
#include "stmt/if.h"
#include "stmt/ifelse.h"
#include "stmt/for.h"
#include "stmt/forever.h"
#include "stmt/repeat.h"
#include "stmt/while.h"
#include "stmt/case.h"
#include "stmt/ssetup.h"

extern symbol_table symboltable;

// Lvalue for blocking assignment.

setup_lvalue::setup_lvalue(scopelist_type &s)
	: _scope(s)
	{}

void
setup_lvalue::operator()(lvalue *l) const
	{
	// Setup of RangeId for index of variable.
	l->_ident->sim_setup(setup_expr(_scope));

	// Setup for range_id should have tested for valid index.  So
	// no need check.
	node_type *node = symboltable.get(l->_ident->index());

	// Lvalues are wires, registers, and function names.
	if (node->get_net() == 0 && node->get_function() == 0)
		{
		strstream buf;
		buf << *l << '(' << node->type_name() << ')';
		sim_seterror(SE_NLVAL, l->_filename, l->_lineno);
		sim_perror(buf);
		}
	}


// Case item for case statement.

setup_case_item::setup_case_item(scopelist_type &s)
	: _scope(s)
	{}

void
setup_case_item::operator()(case_item *c) const
	{
	c->_stmt->sim_setup(setup_stmt(_scope));
	// Default cases do not have expression lists.
	if (c->is_default())
		return;
	case_item::expr_list::iterator itp(c->_expr->begin());
	case_item::expr_list::iterator stop(c->_expr->end());
	for (; itp != stop; ++itp)
		(*itp)->sim_setup(setup_expr(_scope));
	}


// All statements.

setup_stmt::setup_stmt(scopelist_type &s, event_ptr *e)
	: _scope(s)
	{ _event = e; }

void
setup_stmt::operator()(systask_finish *p) const
	{
	task_type *t = new task_type(p->_name, p);
	// System tasks always have global scope.
	t->scope(0);
	task_type::hash_type hv(symboltable.add(t));
	if (hv._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM) hv._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	// System tasks do not need to be setup.
	t->_done_setup = true;
	}

void
setup_stmt::operator()(systask_write *p) const
	{
	task_type *t = new task_type(p->_name, p);
	// System tasks always have global scope.
	t->scope(0);
	task_type::hash_type hv(symboltable.add(t));
	if (hv._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM) hv._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	// System tasks do not need to be setup.
	t->_done_setup = true;
	}

void
setup_stmt::operator()(systask_monitor *p) const
	{
	task_type *t = new task_type(p->_name, p);
	// System tasks always have global scope.
	t->scope(0);
	task_type::hash_type hv(symboltable.add(t));
	if (hv._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM) hv._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	// System tasks do not need to be setup.
	t->_done_setup = true;
	}

void
setup_stmt::operator()(sysfunc_time *p) const
	{
	func_type *f = new func_type(p->_name, p);
	// System functions always have global scope.
	f->scope(0);
	f->_storage = func_type::num_type(31, 0);
	// Need to save index to get the symbol table node, to store
	// the return value (time).  This is a function call.
	p->_index = symboltable.add(f);
	if (p->_index._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM)p->_index._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	// System functions do not need to be setup.
	f->_done_setup = true;
	}

void
setup_stmt::operator()(null_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	}

void
setup_stmt::operator()(seq_block_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	seq_block_stmt::stmt_list::iterator itp(p->_stmt->begin());
	seq_block_stmt::stmt_list::iterator stop(p->_stmt->end());
	for (; itp != stop; ++itp)
		(*itp)->sim_setup(setup_stmt(_scope));
	}

void
setup_stmt::operator()(task_enable_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	// Check for symbol, and save the index.
	p->_index = symboltable.find(p->_name, _scope);
	if (p->_index._scope < 0)
		{
		sim_seterror((SIM_ERR_NUM)p->_index._value, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}

	// Make sure it's a task.
	task_type *node = symboltable.get(p->_index)->get_task();
	if (node == 0)
		{
		sim_seterror(SE_NTASK, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}
	// Tasks are not setup until the task is enabled.  This is a
	// task enable, so setup...
	node->sim_setup(setup_st_node(_scope, 0, p->_argument));

	// Now check argument list.
	if ((p->_name != "$write" && p->_name != "$monitor") &&
	    !check_arguments(p->_argument, node->_iovariable))
		{
		sim_seterror(SE_NARGLST, p->_filename, p->_lineno);
		sim_perror(p->_name);
		}

	// Everything is ok, so setup the argument list, if any.
	if (p->_argument == 0)
		return;
	task_enable_stmt::arg_list::iterator itp(p->_argument->begin());
	task_enable_stmt::arg_list::iterator stop(p->_argument->end());
	if (p->_name == "$write" || p->_name == "$monitor")
		{
		// Output type system task needs special checking...
		SIM_ERR_NUM err(SE_NONE);
		qouted_str *qs = (*itp)->get_qouted_str();
		if (qs == 0)
			err = SE_NOFMT;
		else
			err = check_format(*qs);
		if (err != SE_NONE)
			{
			sim_seterror(err, p->_filename, p->_lineno);
			sim_perror(p->_name);
			}
		++itp;
		}
	// If a $monitor task enable, need to append an event to the
	// expressions in the arguments.
	event_cache_type *cache = 0;
	event_type *ev = 0;
	counted_ptr<event_type> *eptr = 0;
	bool tst = (p->_name == "$monitor");
	if (tst)
		{
		cache = new event_cache_type(p);
		ev = new change_event<stmt_type>(cache, DC);
		eptr = new counted_ptr<event_type>(ev);
		}
	for (; itp != stop; ++itp)
		{
		(*itp)->sim_setup(setup_expr(_scope));
		if (!(*itp)->sim_monitor(monitor_expr(eptr)))
			{
			strstream buf;
			buf << *(*itp);
			sim_seterror(SE_TYPE, (*itp)->_filename, (*itp)->_lineno);
			sim_perror(buf);
			}
		}
	if (tst)
		delete eptr;
	}

void
setup_stmt::operator()(block_assign_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_lval->sim_setup(setup_lvalue(_scope));
	p->_rval->sim_setup(setup_expr(_scope));
	// For continuous assignments, we need to append an event to the
	// variables of the right-hand-side.
	if (_event != 0)
		{
		if (!p->_rval->sim_monitor(monitor_expr(_event)))
			{
			strstream buf;
			buf << *p->_rval;
			sim_seterror(SE_TYPE, p->_filename, p->_lineno);
			sim_perror(buf);
			}
		}
	}

void
setup_stmt::operator()(non_block_assign_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_lval->sim_setup(setup_lvalue(_scope));
	p->_rval->sim_setup(setup_expr(_scope));
	}

void
setup_stmt::operator()(if_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_expr->sim_setup(setup_expr(_scope));
	p->_stmt->sim_setup(setup_stmt(_scope));
	}

void
setup_stmt::operator()(if_else_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_expr->sim_setup(setup_expr(_scope));
	p->_true_stmt->sim_setup(setup_stmt(_scope));
	p->_false_stmt->sim_setup(setup_stmt(_scope));
	}

void
setup_stmt::operator()(for_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_preassign->sim_setup(setup_stmt(_scope));
	p->_expr->sim_setup(setup_expr(_scope));
	p->_postassign->sim_setup(setup_stmt(_scope));
	p->_stmt->keep_dec(true); // Need to set before calling setup.
	p->_stmt->sim_setup(setup_stmt(_scope));
	}

void
setup_stmt::operator()(forever_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_stmt->keep_dec(true); // Need to set before calling setup.
	p->_stmt->sim_setup(setup_stmt(_scope));
  	}

void
setup_stmt::operator()(repeat_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_expr->sim_setup(setup_expr(_scope));
	p->_stmt->keep_dec(true); // Need to set before calling setup.
	p->_stmt->sim_setup(setup_stmt(_scope));
  	}

void
setup_stmt::operator()(while_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_expr->sim_setup(setup_expr(_scope));
	p->_stmt->keep_dec(true); // Need to set before calling setup.
	p->_stmt->sim_setup(setup_stmt(_scope));
  	}

void
setup_stmt::operator()(case_stmt *p) const
	{
	if (p->_dec != 0)
		p->_dec->sim_setup(setup_dec(_scope));
	p->_expr->sim_setup(setup_expr(_scope));
	case_stmt::citem_list::iterator itp(p->_case_item->begin());
	case_stmt::citem_list::iterator stop(p->_case_item->end());
	for (; itp != stop; ++itp)
		{
		(*itp)->sim_setup(setup_case_item(_scope));
		if ((*itp)->is_default() == true)
			{
			if (p->_def == 0)
				{
				case_stmt::citem_list::iterator delme(itp);
				p->_def = *itp;
				--itp; // Move iterator back one.
				p->_case_item->erase(delme);
				}
			else
				{
				sim_seterror(SE_MULTDEF, p->_filename, p->_lineno);
				sim_perror("default:");
				}
			}
		}
	}
