// Verilog Behavioral Simulator
// Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
//
// This file is part of the Verilog Behavioral Simulator package.
// See the file COPYRIGHT for copyright and disclaimer information.
// See the file COPYING for the licensing terms and conditions.
// See the file CONTRIBUTORS for a list of contributing authors.
//
// strigger.cc

#include <stdlib.h> // exit()
#include "common/error.h"
#include "common/time_whl.h"
#include "common/sym_tab.h"
#include "common/st_func.h"
#include "common/st_net.h"
#include "common/st_task.h"
#include "common/st_trig.h"
#include "common/st_util.h"
#include "expr/qstr.h"
#include "expr/eeval.h"
#include "expr/etrigger.h"
#include "misc/dectrig.h"
#include "stmt/lvalue.h"
#include "stmt/finish.h"
#include "stmt/write.h"
#include "stmt/monitor.h"
#include "stmt/time.h"
#include "stmt/seqblk.h"
#include "stmt/taskenbl.h"
#include "stmt/bassign.h"
#include "stmt/nbassign.h"
#include "stmt/if.h"
#include "stmt/ifelse.h"
#include "stmt/for.h"
#include "stmt/forever.h"
#include "stmt/repeat.h"
#include "stmt/while.h"
#include "stmt/case.h"
#include "stmt/strigger.h"

extern symbol_table symboltable;
extern time_wheel<stmt_base> timewheel;
extern event_queue<stmt_base> eventqueue;

// Lvalue for assignment statement.

trigger_lvalue::trigger_lvalue(num_type &n)
	: _num(n) {}

void
trigger_lvalue::operator()(lvalue *l) const
	{
	// Get the storage object from the symbol table.
	node_type *node = symboltable.get(l->_ident->index());
	net_type *net = node->get_net();

	position_type ms, ls;
	select_size_type type = l->_ident->get_select(ms, ls);
	if (net != 0)
		{
		if (type == select_base::SELECT_RANGE)
			net->assignment(_num, ms, ls);
		else if (type == select_base::SELECT_BIT)
			net->assignment(_num, ms);
		else
			net->assignment(_num);
		}
	else
		{
		func_type *func = node->get_function();
		if (type == select_base::SELECT_RANGE)
			func->assignment(_num, ms, ls);
		else if (type == select_base::SELECT_BIT)
			net->assignment(_num, ms);
		else
			func->assignment(_num);
		}
	}


// Case item for case statement.

trigger_case_item::trigger_case_item(num_type &b, bool &s, stmt_type *p)
	: _status(s), _value(b) { _parent = p; }

bool
trigger_case_item::operator()(case_item *p) const
	{
	if (p->is_default())
		return p->_stmt->sim_trigger(trigger_stmt(_parent));
	bool ret(true);
	case_item::expr_list::iterator itp(p->_expr->begin());
	case_item::expr_list::iterator stop(p->_expr->end());
	for (; itp != stop; ++itp)
		{
		if ((*itp)->sim_evaluate(evaluate_expr()) == _value)
			{
			ret = p->_stmt->sim_trigger(trigger_stmt(_parent));
			_status = true;
			break;
			}
		}
	return ret;
	}


// All statements.

trigger_stmt::trigger_stmt(stmt_type *s)
	{ _parent = s; }

// The following few function definitions are system tasks and functions.
// They are not regular statements, so seperate them.

bool
trigger_stmt::operator()(systask_finish *) const
	{
	// systask_finish is an internal statement.  It can not have a
	// delay or event control.  So, just trigger it.
	cout << endl << "Ending simulation at time unit "
		<< timewheel.current_time()
		<< " due to $finish called."
		<< endl << endl;
	sim_errno = SE_NONE;
	exit(sim_errno);
	}

bool
trigger_stmt::operator()(systask_write *) const
	{
	task_caller_type *caller = _parent->get_task_enable();
	// The $write task enable should at least have a format string.
	// No string means nothing to do.
	if (caller->_argument == 0)
		return true;

	// Now, grab the format string and find out what to do.
	task_caller_type::arg_list::iterator itp(caller->_argument->begin());
	task_caller_type::arg_list::iterator stop(caller->_argument->end());
	// Should not be null.  Setup should have checked.
	const str_type &fmt = *((*itp)->get_qouted_str());
	++itp;
	for (size_t i = 1; i < fmt.size(); ++i)
		{
		if (fmt[i] == '\"')
			break;
		switch(fmt[i])
			{
			case '%':
				{
				str_type s;
				++i;	// Move to next character.
				find_format(i, fmt, s);
				if (itp != stop)
					(*itp)->sim_trigger(trigger_expr(s));
				++itp;
				break;
				}
			case '\\':
				// Escape character found.
				print_escapecode(fmt, ++i);
				break;
			default:
				// Other junk gets printed.
				cout << fmt[i];
				break;
			}
		}
	return true;
	}

bool
trigger_stmt::operator()(systask_monitor *) const
	{
	task_caller_type *caller = _parent->get_task_enable();
	// The $monitor task enable should at least have a format string.
	// No string means nothing to do.
	if (caller->_argument == 0)
		return true;

	// Now, grab the format string and find out what to do.
	task_caller_type::arg_list::iterator itp(caller->_argument->begin());
	task_caller_type::arg_list::iterator stop(caller->_argument->end());
	// Should not be null.  Setup should have checked.
	const str_type &fmt = *((*itp)->get_qouted_str());
	++itp;
	for (size_t i = 1; i < fmt.size(); ++i)
		{
		if (fmt[i] == '\"')
			break;
		switch(fmt[i])
			{
			case '%':
				{
				string s;
				++i;	// Move to next character.
				find_format(i, fmt, s);
				if (itp != stop)
					(*itp)->sim_trigger(trigger_expr(s));
				++itp;
				break;
				}
			case '\\':
				// Escape character found.
				print_escapecode(fmt, ++i);
				break;
			default:
				// Other junk gets printed.
				cout << fmt[i];
				break;
			}
		}
	return true;
	}

bool
trigger_stmt::operator()(sysfunc_time *p) const
	{
	// All we need to do is store the current time.
	func_type *func = symboltable.get(p->_index)->get_function();
	func->_storage = timewheel.current_time();
	return true;
	}

// The rest of the function definitions are for statement trigger.

bool
trigger_stmt::operator()(null_stmt *p) const
	{
	// This function is used as a template for all the other statement
	// trigger functions.  Since all of them are the same, no need to
	// recomment each one.

	// Do we need to delay?
	if (p->_dec != 0)
		{
		if (_parent->_always)
			// We are part of an 'always' procedural block.  So
			// append ourselves for next time, but we still need
			// to trigger the action.
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				// We were not asked to persist, so append to future
				// and remove dec.
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			// We are asked to persist, so indicate a state change,
			// and continue.
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	// We should now do whatever is necessary.  For null statements,
	// there is nothing to do, so return.
	return true;
	}

bool
trigger_stmt::operator()(seq_block_stmt *p) const
	{
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}
	// Delays has been handled, so trigger this statement.

	// This is the new simulation procedure.  We must handle one statement
	// at a time.  This means we do not go through the entire list
	// and append each statement to the future.  Instead, we get one
	// statement and try and execute it.  If there's a delay, we append
	// this entire statement into the future (else case in above).

	// Note:  We must push the list of statements into the stack if it's
	// not already on the stack.
	seq_block_stmt *sqblk = p;
	if (!p->_pushed)
		{
#if defined(DEBUG_NESTED)
cout << "DEBUG_NESTED:  Adding list.." << endl;
printptrlist<Stmts *> xx(p->_stmt);
cout << endl;
#endif
		// Set currrent statement to be used later.  This works
		// because <_parent> could also be <seqblk> if we are
		// part of a procedural block.
		p->_curstmt = p->_stmt->begin();
		sqblk->_stk.push(p);
		p->_pushed = true;
		}

	// Let's begin.
	seq_block_stmt *nesting;
	while (true)
		{
		// First, get next statement.
		while (true)
			{
			// Do we have anything to do?
			if (sqblk->_stk.size() < 1)
				return true;
			// Grab the block to execute.
			nesting = sqblk->_stk.top();
			// Check if we came to the end.
			if (nesting->_curstmt != nesting->_stmt->end())
				{
				// We got a statement.
#if defined(DEBUG_NESTED)
cout << "DEBUG_NESTED:  Using list...";
printptrlist<Stmts *> yy(nesting->_stmt);
cout << endl;
#endif
				break;
				}
			else
				{
				// No more statements in this nested level.
#if defined(DEBUG_NESTED)
cout << "DEBUG_NESTED:  Ended one nesting level..." << endl;
#endif
				// Dump this block.
				sqblk->_stk.pop();
				// Reset <pushed>, because we might be
				// triggered again.
				nesting->_pushed = false;
				// If in a loop, need to do extra stuff.
				// So return and do them.
				if (nesting->_dec != 0)
					{
					// Here we use <persist> as an indicator of
					// being in a loop.
					if (nesting->_dec->_persist)
						return true;
					}
				// Otherwise, find another statement.
				}
			}

		// Now trigger this statement.  Let the statement handle
		// the delay.
		stmt_type *tmp = (*nesting->_curstmt).get();
#if defined(DEBUG_NESTED)
cout << "DEBUG_NESTED:  Triggering, " << *tmp << endl;
#endif
		if (tmp->sim_trigger(trigger_stmt(_parent)) == false)
			{
			// Execution has been delayed.
#if defined(DEBUG_NESTED)
cout << "DEBUG_NESTED:  Delayed * level, " << *tmp << endl;
#endif
			return false;
			}
		++(nesting->_curstmt); // Setup for next round.
		}
	}

bool
trigger_stmt::operator()(task_enable_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	// Due to nesting levels, we need to return status.
	task_type *t = symboltable.get(p->_index)->get_task();
	return t->sim_trigger(trigger_st_node(p));
	}

bool
trigger_stmt::operator()(block_assign_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	// Evaluate the expr and store it.
	num_type res(p->_rval->sim_evaluate(evaluate_expr()));
	p->_lval->sim_trigger(trigger_lvalue(res));
	return true;
	}

bool
trigger_stmt::operator()(non_block_assign_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	if (p->_value == 0)
		{
		p->_value = new num_type;
		*(p->_value) = p->_rval->sim_evaluate(evaluate_expr());
		if (!p->_event->is_queued()
		 && (p->_event->type() & TRIG_CHNG) != TRIG_NONE)
			eventqueue.sim_add_event(p->_event);
		}
	else
		{
		p->_lval->sim_trigger(trigger_lvalue(*p->_value));
		delete p->_value;
		p->_value = 0;
		}
	return true;
	}

bool
trigger_stmt::operator()(if_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	if (p->_expr->sim_evaluate(evaluate_expr()) != false)
		return p->_stmt->sim_trigger(trigger_stmt(_parent));
	// Expression evaluated to false, but the 'if' statement is
	// considered executed. Thus, return true.
	return true;
	}

bool
trigger_stmt::operator()(if_else_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	if (p->_expr->sim_evaluate(evaluate_expr()) != false)
		return p->_true_stmt->sim_trigger(trigger_stmt(_parent));
	return p->_false_stmt->sim_trigger(trigger_stmt(_parent));
	}

bool
trigger_stmt::operator()(for_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	// Test whether we were delayed before.
	if (p->_delayed == true)
		{
		// If so, we need to reset the flag and jump into the loop.
		p->_delayed = false;
		// I don't like goto's either, need to find a better way.
		goto continue_from_delay;
		}
	// Begin work.
	p->_preassign->sim_trigger(trigger_stmt(_parent));
	while (p->_expr->sim_evaluate(evaluate_expr()) != false)
		{
continue_from_delay:
		if (p->_stmt->sim_trigger(trigger_stmt(_parent)) == false)
			{
			// A delay in a for loop?  Hope this works.
			p->_delayed = true;
			return false;
			}
		p->_postassign->sim_trigger(trigger_stmt(_parent));
		}
	return true;
	}

bool
trigger_stmt::operator()(forever_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	if (p->_delayed == true)
		{
		p->_delayed = false;
		goto continue_from_delay;
		}
	while (true)
		{
continue_from_delay:
		if (p->_stmt->sim_trigger(trigger_stmt(_parent)) == false)
			{
			p->_delayed = true;
			return false;
			}
		}
	return true;
	}

bool
trigger_stmt::operator()(repeat_stmt *p) const
	{
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	if (p->_delayed == true)
		{
		p->_delayed = false;
		goto continue_from_delay;
		}
	p->_repeat_count = p->_expr->sim_evaluate(evaluate_expr());
	while (p->_repeat_count != false)
		{
continue_from_delay:
		if (p->_stmt->sim_trigger(trigger_stmt(_parent)) == false)
			{
			p->_delayed = true;
			return false;
			}
		--(p->_repeat_count);
		}
	return true;
	}

bool
trigger_stmt::operator()(while_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	if (p->_delayed == true)
		{
		p->_delayed = false;
		goto continue_from_delay;
		}
	while (p->_expr->sim_evaluate(evaluate_expr()) != false)
		{
continue_from_delay:
		if (p->_stmt->sim_trigger(trigger_stmt(_parent)) == false)
			{
			p->_delayed = true;
			return false;
			}
		}
	return true;
	}

bool
trigger_stmt::operator()(case_stmt *p) const
	{
	// See above for comments.
	if (p->_dec != 0)
		{
		if (_parent->_always)
			p->_dec->sim_trigger(trigger_dec(_parent));
		else
			{
			if (p->_dec->_persist == false)
				{
				p->_dec->sim_trigger(trigger_dec(_parent));
				delete p->_dec;
				p->_dec = 0;
				return false;
				}
			if (p->_dec->_delayed_loop == true)
				p->_dec->_delayed_loop = false;
			else
				{
				p->_dec->_delayed_loop = true;
				p->_dec->sim_trigger(trigger_dec(_parent));
				return false;
				}
			}
		}

	bool ret(true);
	bool status(false);
	// Get the expression result first.
	num_type value(p->_expr->sim_evaluate(evaluate_expr()));
	case_stmt::citem_list::iterator itp(p->_case_item->begin());
	case_stmt::citem_list::iterator stop(p->_case_item->end());
	// Find the case item to trigger.
	for (; itp != stop; ++itp)
		{
		ret = (*itp)->sim_trigger(trigger_case_item(value, status, _parent));
		if (status != false)
			break;
		}
	// See if something happened, if not, trigger the default case.
	if (p->_def != 0 && status == false)
		ret = p->_def->sim_trigger(trigger_case_item(value, status, _parent));
	return ret;
	}
