%{
/*
 * Verilog Behavioral Simulator
 * Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
 *
 * This file is part of the Verilog Behavioral Simulator package.
 * See the file COPYRIGHT for copyright and disclaimer information.
 * See the file COPYING for the licensing terms and conditions.
 * See the file CONTRIBUTORS for a list of contributing authors.
 *
 * vbs.l
 *
 * Verilog Behavioral Simulator lexical scanner.
 */

#include <stdio.h> /* fileno(flex) */
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* isatty(flex) */
#include "parser.h"
#include "p_misc.h"
#include "p_moditm.h"
#include "vbs_yacc.h"

extern void copy_word(char *str);
extern void save_word(char *str);
int cur_lineno = 1;
char *cur_filename = 0;
char *savedword = 0;

%}

%s COMMENT

NL           [\n]
WhiteSpace   [\t\r\b{NL} ]
Digit        [0-9]
DigitU       [0-9_]
Letter       [a-zA-Z]
LetterU      [a-zA-Z_]
WordNum      [0-9a-zA-Z]
WordNumU     [0-9a-zA-Z_]
WhiteSpaces  {WhiteSpace}+
Number       {Digit}{DigitU}*
Word         {LetterU}{WordNumU}*
Binary       ({Number})?'[bB][01xXzZ?][01xXzZ?_]*
Octal        ({Number})?'[oO][0-7xXzZ?][0-7xXzZ?_]*
Decimal      ({Number})?'[dD]{Number}
Hexdecimal   ({Number})?'[hH][0-9a-fA-FxXzZ?][0-9a-fA-FxXzZ?_]*
QoutedString \"[^"\n]*["{NL}]
Comment      "//"[^\n]*{NL}

%%

<INITIAL>{NL}           { cur_lineno++; continue; }
<INITIAL>{WhiteSpaces}  { if (yytext[0] == '\n') cur_lineno++; continue; }
<INITIAL>{Comment}      { cur_lineno++; continue; }
<INITIAL>"/*"           { BEGIN COMMENT; continue; }
<COMMENT>{NL}           { cur_lineno++; continue; }
<COMMENT>.              { continue; }
<COMMENT>"*/"           { BEGIN INITIAL; continue; }

<INITIAL>">="           { save_word(">="); return YYGRTEQ; }
<INITIAL>"<="           { save_word("<="); return YYLESEQ; }
<INITIAL>"=="           { save_word("=="); return YYEQEQ; }
<INITIAL>"!="           { save_word("!="); return YYNOTEQ; }
<INITIAL>"||"           { save_word("||"); return YYOROR; }
<INITIAL>"&&"           { save_word("&&"); return YYANDAND; }
<INITIAL>"<<"           { save_word("<<"); return YYLSHIFT; }
<INITIAL>">>"           { save_word(">>"); return YYRSHIFT; }
<INITIAL>always         { save_word("always"); return YYALWAYS; }
<INITIAL>assign         { save_word("assign"); return YYASSIGN; }
<INITIAL>begin          { save_word("begin"); return YYBEGIN; }
<INITIAL>case           { save_word("case"); return YYCASE; }
<INITIAL>default        { save_word("default"); return YYDEFAULT; }
<INITIAL>else           { save_word("else"); return YYELSE; }
<INITIAL>end            { save_word("end"); return YYEND; }
<INITIAL>endcase        { save_word("endcase"); return YYENDCASE; }
<INITIAL>endfunction    { save_word("endfunction"); return YYENDFUNCTION; }
<INITIAL>endmodule      { save_word("endmodule"); return YYENDMODULE; }
<INITIAL>endtask        { save_word("endtask"); return YYENDTASK; }
<INITIAL>for            { save_word("for"); return YYFOR; }
<INITIAL>forever        { save_word("forever"); return YYFOREVER; }
<INITIAL>function       { save_word("function"); return YYFUNCTION; }
<INITIAL>if             { save_word("if"); return YYIF; }
<INITIAL>initial        { save_word("initial"); return YYINITIAL; }
<INITIAL>inout          { save_word("inout"); return YYINOUT; }
<INITIAL>input          { save_word("input"); return YYINPUT; }
<INITIAL>module         { save_word("module"); return YYMODULE; }
<INITIAL>negedge        { save_word("negedge"); return YYNEGEDGE; }
<INITIAL>or             { save_word("or"); return YYOR; }
<INITIAL>output         { save_word("output"); return YYOUTPUT; }
<INITIAL>posedge        { save_word("posedge"); return YYPOSEDGE; }
<INITIAL>reg            { save_word("reg"); return YYREG; }
<INITIAL>repeat         { save_word("repeat"); return YYREPEAT; }
<INITIAL>task           { save_word("task"); return YYTASK; }
<INITIAL>while          { save_word("while"); return YYWHILE; }
<INITIAL>wire           { save_word("wire"); return YYWIRE; }
<INITIAL>\${Word}       { copy_word(yytext); return YYSYSIDENT; }
<INITIAL>{Binary}       { copy_word(yytext); return YYNUMBER; }
<INITIAL>{Octal}        { copy_word(yytext); return YYNUMBER; }
<INITIAL>{Decimal}      { copy_word(yytext); return YYNUMBER; }
<INITIAL>{Hexdecimal}   { copy_word(yytext); return YYNUMBER; }
<INITIAL>{Number}       { copy_word(yytext); return YYNUMBER; }
<INITIAL>{QoutedString} { copy_word(yytext); return YYQSTRING; }
<INITIAL>{Word}         { copy_word(yytext); return YYIDENT; }
<INITIAL>.              { return(yytext[0]); }

%%

void
copy_word(char *str)
	{
	yylval.ident = my_strdup(str);
	if (savedword != 0)
		{
		free(savedword);
		savedword = 0;
		}
	}

void
save_word(char *str)
	{
	if (savedword != 0)
		free(savedword);
	savedword = my_strdup(str);
	}
