%{
/*
 * Verilog Behavioral Simulator
 * Copyright (C) 1995 Lay Hoon Tho, Jimen Ching
 *
 * This file is part of the Verilog Behavioral Simulator package.
 * See the file COPYRIGHT for copyright and disclaimer information.
 * See the file COPYING for the licensing terms and conditions.
 * See the file CONTRIBUTORS for a list of contributing authors.
 *
 * vbs.y
 *
 * Verilog Behavioral Simulator yacc grammar parser.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include "parser.h"
#include "p_expr.h"
#include "p_stmt.h"
#include "p_misc.h"
#include "p_moditm.h"

%}

%union
	{
	char *ident;

	p_module module;
	p_port port;
	p_portlist portlist;
	p_moduleitem moduleitem;
	p_moduleitemlist moduleitemlist;

	/* Module Items. */
	p_initial initial;
	p_always always;
	p_regdecl regdecl;
	p_wiredecl wiredecl;
	p_inputdecl inputdecl;
	p_outputdecl outputdecl;
	p_inoutdecl inoutdecl;
	p_tfdecl tfdecl;
	p_tfdecllist tfdecllist;
	p_function function;
	p_task task;
	p_portconn portconn;
	p_portconnlist portconnlist;
	p_modinstance modinstance;
	p_modinstancelist modinstancelist;
	p_modinstan modinstan;
	p_contassign contassign;
	p_blkassignlist blkassignlist;

	/* Delay or events. */
	p_evntexpr evntexpr;
	p_evntexprlist evntexprlst;
	p_delayevntctl dec;

	/* Statements. */
	p_stmts stmts;
	p_stmtslist stmtslist;
	p_sequentialblock seqblk;
	p_taskstmt systask;
	p_taskstmt taskstmt;
	p_blkassignstmt blkassign;
	p_nblkassignstmt nblkassign;
	p_lvalue lval;
	p_caseitem caseitem;
	p_caseitemlist caseitemlst;

	/* Expressions. */
	p_expression expr;
	p_expressionlist exprlst;
	p_constexpr constexpr;
	p_bit bit;
	p_rangeid rangeid;
	p_range range;
	p_rangeidlist rangeidlst;
	p_qstring qstr;
	p_funcenable funcenable;
	p_number num;
	};

%token YYALWAYS
%token YYASSIGN
%token YYANDAND
%token YYBEGIN
%token YYCASE
%token YYDEFAULT
%token YYELSE
%token YYEND
%token YYENDCASE
%token YYENDFUNCTION
%token YYENDMODULE
%token YYENDTASK
%token YYEQEQ
%token YYFOR
%token YYFOREVER
%token YYFUNCTION
%token YYGRTEQ
%token YYIF
%token YYINITIAL
%token YYINOUT
%token YYINPUT
%token YYLESEQ
%token YYLSHIFT
%token YYNEGEDGE
%token YYNOTEQ
%token YYOR
%token YYOROR
%token YYOUTPUT
%token YYPOSEDGE
%token YYREPEAT
%token YYRSHIFT
%token YYMODULE
%token YYREG
%token YYTASK
%token YYWHILE
%token YYWIRE
%token <ident> YYIDENT /* symbol name (identifier) */
%token <ident> YYNUMBER /* number, all bases */
%token <ident> YYQSTRING /* qouted string, i.e. "abc" */
%token <ident> YYSYSIDENT /* system task or function enable */

%left YYOR
%left YYOROR
%left YYANDAND
%left '|'
%left '&'
%left '^'
%left YYEQEQ YYNOTEQ
%left YYGRTEQ YYLESEQ '>' '<'
%left YYLSHIFT YYRSHIFT
%left '+' '-'
%right '~' '!'
%nonassoc LT_YYELSE
%nonassoc YYELSE

%type <module> module
%type <ident> name_of_module
%type <portlist> list_of_ports_question
%type <portlist> port_plus
%type <port> port
%type <ident> name_of_port

%type <dec> delay_or_event_cntl
%type <dec> delay_control
%type <dec> event_control
%type <evntexpr> event_expression
%type <evntexprlst> ored_event_expression

%type <moduleitemlist> module_item_star
%type <moduleitem> module_item
%type <initial> initial_block
%type <always> always_block
%type <regdecl> reg_declaration
%type <wiredecl> wire_declaration
%type <inputdecl> input_declaration
%type <outputdecl> output_declaration
%type <inoutdecl> inout_declaration
%type <tfdecl> tf_declaration
%type <tfdecllist> tf_declaration_star
%type <tfdecllist> tf_declaration_plus
%type <function> function_definition
%type <task> task_definition
%type <portconnlist> module_connections_plus
%type <portconnlist> module_connections_question
%type <portconn> module_port_connection
%type <ident> name_of_instance
%type <modinstance> module_instance
%type <modinstancelist> module_instance_plus
%type <modinstan> module_instantiation
%type <dec> delay_control_question
%type <blkassignlist> block_assign_plus
%type <contassign> continuous_assign

%type <stmtslist> statement_star
%type <stmts> statement_or_null
%type <stmts> statement
%type <seqblk> sequential_block
%type <systask> system_task_enable
%type <taskstmt> task_enable
%type <ident> name_of_system_task
%type <ident> name_of_task
%type <blkassign> blocking_assign
%type <nblkassign> non_blocking_assign
%type <lval> lvalue
%type <ident> identifier
%type <caseitem> case_item
%type <caseitemlst> case_item_plus

%type <exprlst> expression_plus
%type <expr> expression
%type <expr> primary
%type <constexpr> const_expression
%type <bit> bit
%type <rangeidlst> list_of_registers
%type <rangeid> name_of_register
%type <range> range_question
%type <range> range
%type <num> number

%type <funcenable> function_enable
%type <ident> name_of_system_function
%type <ident> name_of_function

%%

source_text:
		description_star
	;

description_star:
		/* Optional. */
	|	description_star description
	;

description:
		module
			/* Save the module so we can let yywrap handle it. */
			{ sim_parsed_module($1); }
	;

module:
		YYMODULE name_of_module list_of_ports_question ';'
				module_item_star
		YYENDMODULE
			{ $$ = p_create_module($2, $5, $3); }
	;

name_of_module:
		YYIDENT
			{ $$ = $1; }
	;

list_of_ports_question:
		/* Optional. */
			{ $$ = 0; }
	|	'(' port_plus ')'
			{ $$ = $2; }
	;

port_plus:
		port
			{ $$ = p_create_portlist(0, $1); }
	|	port_plus ',' port
			{ $$ = p_create_portlist($1, $3); }
	;

port:
		name_of_port
			{ $$ = p_create_port($1); }
	;

name_of_port:
		YYIDENT
			{ $$ = $1; }
	;

module_item_star:
		/* Zero or more. */
			{ $$ = p_create_moduleitemlist(0, 0); }
	|	module_item_star module_item
			{ $$ = p_create_moduleitemlist($1, $2); }
	;

module_item:
		initial_block
			{ $$ = $1; }
	|	always_block
			{ $$ = $1; }
	|	reg_declaration
			{ $$ = $1; }
	|	wire_declaration
			{ $$ = $1; }
	|	input_declaration
			{ $$ = $1; }
	|	output_declaration
			{ $$ = $1; }
	|	inout_declaration
			{ $$ = $1; }
	|	function_definition
			{ $$ = $1; }
	|	task_definition
			{ $$ = $1; }
	|	module_instantiation
			{ $$ = $1; }
	|	continuous_assign
			{ $$ = $1; }
	;

initial_block:
		YYINITIAL statement
			{ $$ = p_create_initial($2); }
	;

always_block:
		YYALWAYS statement
			{ $$ = p_create_always($2); }
	;

reg_declaration:
		YYREG range_question list_of_registers ';'
			{ $$ = p_create_regdecl($3, $2); }
	;

wire_declaration:
		YYWIRE range_question list_of_registers ';'
			{ $$ = p_create_wiredecl($3, $2); }
	;

input_declaration:
		YYINPUT range_question list_of_registers ';'
			{ $$ = p_create_inputdecl($3, $2); }
	;

output_declaration:
		YYOUTPUT range_question list_of_registers ';'
			{ $$ = p_create_outputdecl($3, $2); }
	;

inout_declaration:
		YYINOUT range_question list_of_registers ';'
			{ $$ = p_create_inoutdecl($3, $2); }
	;

function_definition:
		YYFUNCTION range_question name_of_function ';'
				tf_declaration_plus
				statement
		YYENDFUNCTION
			{ $$ = p_create_function($3, $6, $5, $2); }
	;

task_definition:
		YYTASK name_of_task ';'
				tf_declaration_star
				statement_or_null
		YYENDTASK
			{ $$ = p_create_task($2, $5, $4); }
	;

continuous_assign:
		YYASSIGN delay_control_question block_assign_plus ';'
			{ $$ = p_create_contassign($3, $2); }
	;

delay_control_question:
		/* Optional. */
			{ $$ = 0; }
	|	delay_control
			{ $$ = $1; }
	;

block_assign_plus:
		blocking_assign
			{ $$ = p_create_blkassignlist(0, $1); }
	|	block_assign_plus ',' blocking_assign
			{ $$ = p_create_blkassignlist($1, $3); }
	;

tf_declaration_plus:
		tf_declaration
			{ $$ = p_create_tfdecllist(0, $1); }
	|	tf_declaration_plus tf_declaration
			{ $$ = p_create_tfdecllist($1, $2); }
	;

tf_declaration_star:
		/* Optional. */
			{ $$ = p_create_tfdecllist(0, 0); }
	|	tf_declaration_star tf_declaration
			{ $$ = p_create_tfdecllist($1, $2); }
	;

tf_declaration:
		reg_declaration
			{ $$ = $1; }
	|	input_declaration
			{ $$ = $1; }
	|	output_declaration
			{ $$ = $1; }
	|	inout_declaration
			{ $$ = $1; }
	;

module_instantiation:
		name_of_module module_instance_plus ';'
			{ $$ = p_create_modinstan($1, $2); }
	;

module_instance_plus:
		module_instance
			{ $$ = p_create_modinstancelist(0, $1); }
	|	module_instance_plus ',' module_instance
			{ $$ = p_create_modinstancelist($1, $3); }
	;

module_instance:
		name_of_instance '(' module_connections_question ')'
			{ $$ = p_create_modinstance($1, $3); }
	;

name_of_instance:
		YYIDENT
			{ $$ = $1; }
	;

module_connections_question:
		/* Optional. */
			{ $$ = 0; }
	|	module_connections_plus
			{ $$ = $1; }
	;

module_connections_plus:
		module_port_connection
			{ $$ = p_create_portconnlist(0, $1); }
	|	module_connections_plus ',' module_port_connection
			{ $$ = p_create_portconnlist($1, $3); }
	;

module_port_connection:
		expression
			{ $$ = p_create_portconn($1); }
/*	|	*//* NULL */
			/*
			 * The module_connections_question above has an empty
			 * rule, thus this empty rule will cause a reduce/reduce
			 * error from yacc.  This rule is removed because it
			 * seems to have an easier work around in Verilog.
			 */
/*			{ $$ = p_create_portconn(0); }*/
	;

range_question:
		/* Optional. */
			{ $$ = 0; }
	|	range
			{ $$ = $1; }
	;

range:
		'[' const_expression ':' const_expression ']'
			{ $$ = p_create_range($2, $4); }
	;

list_of_registers:
		name_of_register
			{ $$ = p_create_rangeidlist(0, $1); }
	|	list_of_registers ',' name_of_register
			{ $$ = p_create_rangeidlist($1, $3); }
	;

name_of_register:
		YYIDENT
			{ $$ = p_create_rangeid($1, 0); }
	;

statement_star:
		/* Zero or more. */
			{ $$ = p_create_stmtslist(0, 0); }
	|	statement_star statement
			{ $$ = p_create_stmtslist($1, $2); }
	;

statement_or_null:
		statement
			{ $$ = $1; }
	|	';'
			{ $$ = p_create_stmts(); }
	;

statement:
		blocking_assign ';'
			{ $$ = $1; }
	|	non_blocking_assign ';'
			{ $$ = $1; }
	|	delay_or_event_cntl statement_or_null
			{ $$ = p_create_includedec($2, $1); }
	|	sequential_block
			{ $$ = $1; }
	|	system_task_enable
			{ $$ = $1; }
	|	task_enable
			{ $$ = $1; }
	|	YYIF '(' expression ')' statement_or_null %prec LT_YYELSE
			{ $$ = p_create_ifstmt($3, $5); }
	|	YYIF '(' expression ')' statement_or_null YYELSE statement_or_null
			{ $$ = p_create_ifelsestmt($3, $5, $7); }
	|	YYCASE '(' expression ')' case_item_plus YYENDCASE
			{ $$ = p_create_casestmt($3, $5); }
	|	YYFOR '(' blocking_assign ';' expression ';' blocking_assign ')'
			statement
			{ $$ = p_create_forstmt($3, $5, $7, $9); }
	|	YYFOREVER
			statement
			{ $$ = p_create_foreverstmt($2); }
	|	YYREPEAT '(' expression ')'
			statement
			{ $$ = p_create_repeatstmt($3, $5); }
	|	YYWHILE '(' expression ')'
			statement
			{ $$ = p_create_whilestmt($3, $5); }
	;

blocking_assign:
		lvalue '=' expression
			{ $$ = p_create_blkassignstmt($1, $3); }
	;

non_blocking_assign:
		lvalue YYLESEQ expression
			{ $$ = p_create_nblkassignstmt($1, $3); }
	;

lvalue:
		identifier
			{
			p_rangeid tmp = p_create_rangeid($1, 0);
			$$ = p_create_lvalue(tmp);
			}
	|	identifier bit
			{
			p_rangeid tmp = p_create_rangeid($1, $2);
			$$ = p_create_lvalue(tmp);
			}
	|	identifier range
			{
			p_rangeid tmp = p_create_rangeid($1, $2);
			$$ = p_create_lvalue(tmp);
			}
	;

identifier:
		YYIDENT
			{ $$ = $1; }
	;

bit:
		'[' expression ']'
			{ $$ = p_create_bit($2); }
	;

delay_or_event_cntl:
		delay_control
	|	event_control
	;

delay_control:
		'#' number
			{ $$ = p_create_delaynum($2); }
	|	'#' identifier
			{ $$ = p_create_delayid($2); }
	;

event_control:
		'@' identifier
			{ $$ = p_create_evntid($2); }
	|	'@' '(' ored_event_expression ')'
			{ $$ = p_create_orevntexpr($3); }
	;

event_expression:
		expression
			{ $$ = p_create_chngevntexpr($1); }
	|	YYPOSEDGE expression
			{ $$ = p_create_posevntexpr($2); }
	|	YYNEGEDGE expression
			{ $$ = p_create_negevntexpr($2); }
	;

ored_event_expression:
		event_expression
			{ $$ = p_create_evntexprlist(0, $1); }
	|	ored_event_expression YYOR event_expression
			{ $$ = p_create_evntexprlist($1, $3); }
	;

sequential_block:
		YYBEGIN statement_star YYEND
			{ $$ = p_create_sequentialblock($2); }
	;

system_task_enable:
		name_of_system_task ';'
			{ $$ = p_create_taskstmt($1, 0); }
	|	name_of_system_task '(' expression_plus ')' ';'
			{ $$ = p_create_taskstmt($1, $3); }
	;

name_of_system_task:
		YYSYSIDENT
			{ $$ = $1; }
	;

task_enable:
		name_of_task ';'
			{ $$ = p_create_taskstmt($1, 0); }
	|	name_of_task '(' expression_plus ')' ';'
			{ $$ = p_create_taskstmt($1, $3); }
	;

name_of_task:
		YYIDENT
			{ $$ = $1; }
	;

case_item_plus:
		case_item
			{ $$ = p_create_caseitemlist(0, $1); }
	|	case_item_plus case_item
			{ $$ = p_create_caseitemlist($1, $2); }
	;

case_item:
		expression_plus ':' statement_or_null
			{ $$ = p_create_caseitem($1, $3); }
	|	YYDEFAULT ':' statement_or_null
			{ $$ = p_create_caseitem(0, $3); }
	|	YYDEFAULT statement_or_null
			{ $$ = p_create_caseitem(0, $2); }
	;

expression_plus:
		expression
			{ $$ = p_create_expressionlist(0, $1); }
	|	expression_plus ',' expression
			{ $$ = p_create_expressionlist($1, $3); }
	;

expression:
		primary
			{ $$ = $1; }
	|	'~' primary
			{ $$ = p_create_invert($2); }
	|	'!' primary
			{ $$ = p_create_not($2); }
	|	expression '+' expression
			{ $$ = p_create_add($1, $3); }
	|	expression '-' expression
			{ $$ = p_create_subtract($1, $3); }
	|	expression '&' expression
			{ $$ = p_create_logand($1, $3); }
	|	expression '|' expression
			{ $$ = p_create_logor($1, $3); }
	|	expression '^' expression
			{ $$ = p_create_logxor($1, $3); }
	|	expression YYEQEQ expression
			{ $$ = p_create_eqeq($1, $3); }
	|	expression YYNOTEQ expression
			{ $$ = p_create_noteq($1, $3); }
	|	expression YYOROR expression
			{ $$ = p_create_oror($1, $3); }
	|	expression YYANDAND expression
			{ $$ = p_create_andand($1, $3); }
	|	expression YYLSHIFT expression
			{ $$ = p_create_lshift($1, $3); }
	|	expression YYRSHIFT expression
			{ $$ = p_create_rshift($1, $3); }
	|	expression YYGRTEQ expression
			{ $$ = p_create_grteq($1, $3); }
	|	expression '>' expression
			{ $$ = p_create_grtthan($1, $3); }
	|	expression YYLESEQ expression
			{ $$ = p_create_leseq($1, $3); }
	|	expression '<' expression
			{ $$ = p_create_lesthan($1, $3); }
	|	YYQSTRING
			{ $$ = p_create_qstring($1); }
	|	'(' expression ')'
			{ $$ = $2; }
	;

const_expression:
		expression
			{ $$ = p_create_constexpr($1); }
	;

primary:
		number
			{ $$ = $1; }
	|	identifier
			{ $$ = p_create_rangeid($1, 0); }
	|	identifier bit
			{ $$ = p_create_rangeid($1, $2); }
	|	identifier range
			{ $$ = p_create_rangeid($1, $2); }
	|	function_enable
			{ $$ = $1; }
	;

function_enable:
		name_of_system_function
			{
			$$ = p_create_funcenable($1, 0);
			}
	|	name_of_function '(' expression_plus ')'
			{
			$$ = p_create_funcenable($1, $3);
			}
	;

name_of_system_function:
		YYSYSIDENT
			{ $$ = $1; }
	;

name_of_function:
		YYIDENT
			{ $$ = $1; }
	;

number:
		YYNUMBER
			{ $$ = p_create_number($1); }
	;

%%

void
yyerror(char *str)
	{
	if (savedword != 0)
		sim_perror("keyword", str, savedword, cur_filename, cur_lineno);
	else if (yylval.ident != 0 && yylval.ident[0] < 0)
		sim_perror("eof", str, "EOF", cur_filename, cur_lineno);
	else
		/* A symbol is anything besides an EOF. */
		sim_perror("symbol", str, yylval.ident, cur_filename, cur_lineno);
	}
